import os
import glob
import argparse
from Bio import SeqIO


OMA2_usage = '''
============================== OMA2 example commands ==============================

TreeSAK OMA2 -i OrthologousGroups.txt -s OrthologousGroupsFasta -o op_dir -f -n 3
TreeSAK OMA2 -i OrthologousGroups.txt -s OrthologousGroupsFasta -o op_dir -f -c 85

===================================================================================
'''


def sep_path_basename_ext(file_in):

    f_path, f_name = os.path.split(file_in)
    if f_path == '':
        f_path = '.'
    f_base, f_ext = os.path.splitext(f_name)

    return f_name, f_path, f_base, f_ext[1:]


def get_gnm_og_cov(og_dir, og_ext, og_cov_txt):

    og_file_re   = '%s/*.%s' % (og_dir, og_ext)
    og_file_list = glob.glob(og_file_re)

    gnm_to_og_dict = dict()
    for og_file in og_file_list:
        _, _, og_id, _ = sep_path_basename_ext(og_file)
        for each_seq in SeqIO.parse(og_file, 'fasta'):
            seq_id = each_seq.id
            gnm_id = '_'.join(seq_id.split('_')[:-1])
            if gnm_id not in gnm_to_og_dict:
                gnm_to_og_dict[gnm_id] = set()
            gnm_to_og_dict[gnm_id].add(og_id)

    og_cov_txt_handle = open(og_cov_txt, 'w')
    for each_gnm in sorted(list(gnm_to_og_dict.keys())):
        gnm_og_set = gnm_to_og_dict[each_gnm]
        og_cov = len(gnm_og_set)*100/len(og_file_list)
        og_cov = float("{0:.2f}".format(og_cov))
        og_cov_txt_handle.write('%s\t%s\n' % (each_gnm, og_cov))
    og_cov_txt_handle.close()


def get_ortho_to_gene_dict(ortho_groups_txt, og_program):

    ortho_to_gene_dict = dict()
    for each_og in open(ortho_groups_txt):
        if not each_og.startswith('#'):
            og_id = ''
            gene_list = []
            if og_program == 'orthofinder':
                each_og_split = each_og.strip().split(' ')
                og_id = each_og_split[0][:-1]
                gene_list = each_og_split[1:]
            elif og_program == 'oma':
                each_og_split = each_og.strip().split('\t')
                og_id = each_og_split[0]
                group_member_list = each_og_split[1:]
                for each_protein in group_member_list:
                    protein_id = each_protein.split(' ')[0].split(':')[1]
                    gene_list.append(protein_id)
            ortho_to_gene_dict[og_id] = gene_list

    return ortho_to_gene_dict


def select_seq(seq_in, seq_id_list, seq_out):
    output_file_handle = open(seq_out, 'w')
    for seq_record in SeqIO.parse(seq_in, 'fasta'):
        if seq_record.id in seq_id_list:
            output_file_handle.write('>%s\n' % seq_record.id)
            output_file_handle.write('%s\n' % seq_record.seq)
    output_file_handle.close()


def OMA2(args):

    og_txt              = args['i']
    og_seq_dir          = args['s']
    gnm_txt             = args['g']
    op_dir              = args['o']
    force_overwrite     = args['f']
    min_gene_num        = args['n']
    min_gene_cov        = args['c']

    if (min_gene_num is None) and (min_gene_cov is None):
        print('Please specify either -c or -n, program exited!')
        exit()
    elif (min_gene_num is not None) and (min_gene_cov is not None):
        print('-c and -n are not compatible, program exited!')
        exit()

    og_txt_out = ''
    gnm_og_num_txt = ''
    filtered_seq_dir = ''
    if min_gene_num is not None:
        og_txt_out       = '%s/OrthologousGroups_num%s.txt'             % (op_dir, min_gene_num)
        gnm_og_num_txt   = '%s/OrthologousGroups_num%s_per_genome.txt'  % (op_dir, min_gene_num)
        filtered_seq_dir = '%s/OrthologousGroupsFasta_num%s'            % (op_dir, min_gene_num)
    if min_gene_cov is not None:
        og_txt_out       = '%s/OrthologousGroups_cov%s.txt'             % (op_dir, min_gene_cov)
        gnm_og_num_txt   = '%s/OrthologousGroups_cov%s_per_genome.txt'  % (op_dir, min_gene_cov)
        filtered_seq_dir = '%s/OrthologousGroupsFasta_cov%s'            % (op_dir, min_gene_cov)

    # check genome files
    interested_gnm_set = set()
    if gnm_txt is not None:
        if os.path.isfile(gnm_txt) is True:
            for each_gnm in open(gnm_txt):
                gnm_id = each_gnm.strip().split()[0]
                interested_gnm_set.add(gnm_id)
        else:
            print('%s not found, program exited!' % gnm_txt)
            exit()

    # create dir
    if os.path.isdir(op_dir) is True:
        if force_overwrite is True:
            os.system('rm -r %s' % op_dir)
        else:
            print('output folder detected, program exited!')
            exit()
    os.system('mkdir %s' % op_dir)
    os.system('mkdir %s' % filtered_seq_dir)

    # get overall genome set
    overall_gnm_set = set()
    for each_line in open(og_txt):
        if not each_line.startswith('#'):
            each_line_split = each_line.strip().split('\t')
            gene_list = each_line_split[1:]
            for each_gene in gene_list:
                gene_gnm = each_gene.split(':')[0]
                overall_gnm_set.add(gene_gnm)

    qualified_og_set = set()
    id_to_name_dict = dict()
    gene_to_extract_dict = dict()
    og_txt_out_handle = open(og_txt_out, 'w')
    for each_line in open(og_txt):
        if not each_line.startswith('#'):
            each_line_split = each_line.strip().split('\t')
            og_id = each_line_split[0]
            filename = 'OG%s' % int(og_id[3:])
            id_to_name_dict[og_id] = filename
            gene_list = each_line_split[1:]
            filtered_gene_set = set()
            for each_gene in gene_list:
                gene_gnm = each_gene.split(':')[0]
                gene_id = each_gene.split(':')[1].split(' ')[0]
                if len(interested_gnm_set) == 0:
                    filtered_gene_set.add(gene_id)
                else:
                    if gene_gnm in interested_gnm_set:
                        filtered_gene_set.add(gene_id)

            qualified_og = False
            if min_gene_num is not None:
                if len(filtered_gene_set) >= float(min_gene_num):
                    qualified_og = True
            if min_gene_cov is not None:
                if len(interested_gnm_set) == 0:
                    gnm_cov = len(filtered_gene_set)*100/len(overall_gnm_set)
                else:
                    gnm_cov = len(filtered_gene_set)*100/len(interested_gnm_set)

                if gnm_cov >= float(min_gene_cov):
                    qualified_og = True

            if qualified_og is True:
                qualified_og_set.add(og_id)
                og_txt_out_handle.write('%s\t%s\n' % (filename, ','.join(sorted(list(filtered_gene_set)))))
                gene_to_extract_dict[og_id] = filtered_gene_set
    og_txt_out_handle.close()

    for each_og in gene_to_extract_dict:
        seq_file_name = id_to_name_dict[each_og]
        source_file   = '%s/%s.fa' % (og_seq_dir, seq_file_name)
        filtered_file = '%s/%s.fa' % (filtered_seq_dir, seq_file_name)
        select_seq(source_file, gene_to_extract_dict[each_og], filtered_file)

    # get_gnm_og_cov
    get_gnm_og_cov(filtered_seq_dir, 'fa', gnm_og_num_txt)

    # report
    if min_gene_num is not None:
        print('The number of OG with genes >= %s is %s' % (min_gene_num, len(qualified_og_set)))
    if min_gene_cov is not None:
        print('The number of OG with coverage >= %s is %s' % (min_gene_cov, len(qualified_og_set)))

    print('Done!')


if __name__ == '__main__':

    OMA2_parser = argparse.ArgumentParser()
    OMA2_parser.add_argument('-i',   required=True,                         help='OrthologousGroups.txt')
    OMA2_parser.add_argument('-s',   required=True,                         help='sequence dir, OrthologousGroupsFasta')
    OMA2_parser.add_argument('-g',   required=False, default=None,          help='interested genomes')
    OMA2_parser.add_argument('-o',   required=True,  default=None,          help='output directory')
    OMA2_parser.add_argument('-n',   required=False, default=None,          help='minimal number of gene in a OG, not compatible with -c')
    OMA2_parser.add_argument('-c',   required=False, default=None,          help='minimal genome coverage cutoff, not compatible with -n')
    OMA2_parser.add_argument('-f',   required=False, action="store_true",   help='force overwrite')
    args = vars(OMA2_parser.parse_args())
    OMA2(args)



import os.path
import random
import dendropy
import argparse
from ete3 import Tree


RootTreeGTDB226_usage = '''
========================================== RootTreeGTDB226 example command ==========================================   

TreeSAK RootTreeGTDB226 -add_root -d ar -tree ar53.tree -tax ar53.summary.tsv -db db_dir -o ar53.rooted.tree
TreeSAK RootTreeGTDB226 -add_root -d bac -tree bac120.tree -tax bac120.summary.tsv -db db_dir -o bac120.rooted.tree

# Need to download and decompress the following files to your database folder (provide with -db)
https://data.ace.uq.edu.au/public/gtdb/data/releases/release226/226.0/ar53_r226.tree.tar.gz
https://data.ace.uq.edu.au/public/gtdb/data/releases/release226/226.0/bac120_r226.tree.tar.gz
https://data.ace.uq.edu.au/public/gtdb/data/releases/release226/226.0/ar53_metadata_r226.tsv.gz
https://data.ace.uq.edu.au/public/gtdb/data/releases/release226/226.0/bac120_metadata_r226.tsv.gz

=====================================================================================================================
'''


def get_smallest_outgroup(tree_object):

    min_outgroup_leaf_num = 99999
    for each_root_child in tree_object.children:
        leaf_list = each_root_child.get_leaf_names()
        if len(leaf_list) < min_outgroup_leaf_num:
            min_outgroup_leaf_num = len(leaf_list)

    out_group_leaf_list = []
    for each_root_child in tree_object.children:
        leaf_list = each_root_child.get_leaf_names()
        if len(leaf_list) == min_outgroup_leaf_num:
            out_group_leaf_list = leaf_list

    return out_group_leaf_list


def sep_taxon_str(taxon_string):

    taxon_string_split = taxon_string.strip().split(';')
    taxon_p = taxon_string_split[1]
    taxon_c = taxon_string_split[2]
    taxon_o = taxon_string_split[3]
    taxon_f = taxon_string_split[4]
    taxon_g = taxon_string_split[5]

    return taxon_p, taxon_c, taxon_o, taxon_f, taxon_g


def subset_and_rename_tree(tree_file_in, to_keep_leaf_list, rename_dict):

    input_tree = Tree(tree_file_in, quoted_node_names=True, format=1)

    # subset tree
    subset_tree = input_tree.copy()
    subset_tree.prune(to_keep_leaf_list, preserve_branch_length=True)

    # rename leaf
    for each_leaf in subset_tree:
        leaf_name_new = rename_dict.get(each_leaf.name, each_leaf.name)
        each_leaf.name = leaf_name_new

    return subset_tree


def root_with_outgroup(input_tree, out_group_list, add_root_branch, tree_file_rooted):

    """
    Reroot the tree using the given outgroup.
    modified based on: https://github.com/Ecogenomics/GTDBTk/blob/master/gtdbtk/reroot_tree.py

    input_tree:  File containing Newick tree to rerooted.
    output_tree: Name of file for rerooted tree.
    outgroup:    Labels of taxa in outgroup.
    """

    tree = dendropy.Tree.get_from_path(input_tree, schema='newick', rooting='force-rooted', preserve_underscores=True)

    outgroup_in_tree = set()
    ingroup_leaves = set()
    for n in tree.leaf_node_iter():
        if n.taxon.label in out_group_list:
            outgroup_in_tree.add(n.taxon)
        else:
            ingroup_leaves.add(n)

    # Since finding the MRCA is a rooted tree operation, the tree is first rerooted on an ingroup taxa. This
    # ensures the MRCA of the outgroup can be identified so long as the outgroup is monophyletic. If the
    # outgroup is polyphyletic trying to root on it is ill-defined. To try and pick a "good" root for
    # polyphyletic outgroups, random ingroup taxa are selected until two of them give the same size
    # lineage. This will, likely, be the smallest bipartition possible for the given outgroup though
    # this is not guaranteed.

    mrca = tree.mrca(taxa=outgroup_in_tree)
    mrca_leaves = len(mrca.leaf_nodes())
    while True:
        rnd_ingroup = random.sample(list(ingroup_leaves), 1)[0]
        tree.reroot_at_edge(rnd_ingroup.edge, length1=0.5 * rnd_ingroup.edge_length, length2=0.5 * rnd_ingroup.edge_length)
        mrca = tree.mrca(taxa=outgroup_in_tree)
        if len(mrca.leaf_nodes()) == mrca_leaves:
            break
        mrca_leaves = len(mrca.leaf_nodes())

    if mrca.edge_length is not None:
        tree.reroot_at_edge(mrca.edge, length1=0.5 * mrca.edge_length, length2=0.5 * mrca.edge_length)

        # tree.write_to_path(tree_file_rooted, schema='newick', suppress_rooting=True, unquoted_underscores=True)
        tree_out_string = tree.as_string(schema='newick', suppress_rooting=True, unquoted_underscores=True)
        tree_out_string = tree_out_string.replace("'", "")

        # add the root bar
        if add_root_branch is True:
            tree_out_string = '(' + tree_out_string
            tree_out_string = tree_out_string.replace(');', '):0.02);')

        # write out tree string
        tree_file_rooted_handle = open(tree_file_rooted, 'w')
        tree_file_rooted_handle.write(tree_out_string)
        tree_file_rooted_handle.close()


def RootTreeGTDB226(args):

    input_unrooted_tree = args['tree']
    user_gnm_taxon      = args['tax']
    db_dir              = args['db']
    gnm_domain          = args['d']
    add_root_branch     = args['add_root']
    rooted_tree         = args['o']

    missing_file_list = []
    if os.path.isfile(input_unrooted_tree) is False:
        missing_file_list.append(input_unrooted_tree)
    if os.path.isdir(db_dir) is False:
        missing_file_list.append(db_dir)
    if len(missing_file_list) > 0:
        print('Missing files:')
        for missing_file in missing_file_list:
            print(missing_file)
        print('Program exited!')
        exit()

    leaf_list = []
    for leaf in Tree(input_unrooted_tree, format=1):
        leaf_name = leaf.name
        leaf_list.append(leaf_name)

    # define file name
    gtdb_ref_tree_ar    = '%s/ar53_r226.tree'           % db_dir
    gtdb_ref_tree_bac   = '%s/bac120_r226.tree'         % db_dir
    gtdb_gnm_meta_ar    = '%s/ar53_metadata_r226.tsv'   % db_dir
    gtdb_gnm_meta_bac   = '%s/bac120_metadata_r226.tsv' % db_dir

    if gnm_domain == 'bac':
        gtdb_ref_tree = gtdb_ref_tree_bac
        gtdb_gnm_metadata = gtdb_gnm_meta_bac
    elif gnm_domain == 'ar':
        gtdb_ref_tree = gtdb_ref_tree_ar
        gtdb_gnm_metadata = gtdb_gnm_meta_ar
    else:
        print('please provide either "ar" or "bac" to -d')
        exit()

    tree = Tree(gtdb_ref_tree, quoted_node_names=True, format=1)
    ref_tree_gnm_list = tree.get_leaf_names()
    ref_tree_gnm_set = {i for i in ref_tree_gnm_list}

    # read in user_gnm_taxon
    user_gnm_taxon_dict_p = dict()
    user_gnm_taxon_dict_c = dict()
    user_gnm_taxon_dict_o = dict()
    user_gnm_taxon_dict_f = dict()
    user_gnm_taxon_dict_g = dict()
    for each_gnm in open(user_gnm_taxon):
        if not each_gnm.startswith('user_genome\t'):
            each_gnm_split = each_gnm.strip().split('\t')
            gnm_id = each_gnm_split[0]
            gnm_taxon = each_gnm_split[1]

            if gnm_id in leaf_list:
                count_current_gnm = False
                if gnm_domain == 'bac':
                    if 'd__Bacteria' in gnm_taxon:
                        count_current_gnm = True
                elif gnm_domain == 'ar':
                    if 'd__Archaea' in gnm_taxon:
                        count_current_gnm = True

                if count_current_gnm is True:
                    gnm_p, gnm_c, gnm_o, gnm_f, gnm_g = sep_taxon_str(gnm_taxon)

                    if gnm_p not in user_gnm_taxon_dict_p:
                        user_gnm_taxon_dict_p[gnm_p] = set()
                    if gnm_c not in user_gnm_taxon_dict_c:
                        user_gnm_taxon_dict_c[gnm_c] = set()
                    if gnm_o not in user_gnm_taxon_dict_o:
                        user_gnm_taxon_dict_o[gnm_o] = set()
                    if gnm_f not in user_gnm_taxon_dict_f:
                        user_gnm_taxon_dict_f[gnm_f] = set()
                    if gnm_g not in user_gnm_taxon_dict_g:
                        user_gnm_taxon_dict_g[gnm_g] = set()

                    user_gnm_taxon_dict_p[gnm_p].add(gnm_id)
                    user_gnm_taxon_dict_c[gnm_c].add(gnm_id)
                    user_gnm_taxon_dict_o[gnm_o].add(gnm_id)
                    user_gnm_taxon_dict_f[gnm_f].add(gnm_id)
                    user_gnm_taxon_dict_g[gnm_g].add(gnm_id)

    # determine rooting rank, start from phylum
    rooting_rank = ''
    rooting_rank_taxon_dict = dict()
    if len(user_gnm_taxon_dict_p) > 1:
        rooting_rank = 'p'
        rooting_rank_taxon_dict = user_gnm_taxon_dict_p
    elif len(user_gnm_taxon_dict_c) > 1:
        rooting_rank = 'c'
        rooting_rank_taxon_dict = user_gnm_taxon_dict_c
    elif len(user_gnm_taxon_dict_o) > 1:
        rooting_rank = 'o'
        rooting_rank_taxon_dict = user_gnm_taxon_dict_o
    elif len(user_gnm_taxon_dict_f) > 1:
        rooting_rank = 'f'
        rooting_rank_taxon_dict = user_gnm_taxon_dict_f
    elif len(user_gnm_taxon_dict_g) > 1:
        rooting_rank = 'g'
        rooting_rank_taxon_dict = user_gnm_taxon_dict_g

    if rooting_rank == '':
        print('All user genomes are from the same genus, program exited!')
        exit()

    col_index = {}
    canditate_gnms_rooting_rank = dict()
    counted_taxons_rooting_rank = set()
    for each_ref in open(gtdb_gnm_metadata):
            each_ref_split = each_ref.strip().split('\t')
            if each_ref.startswith('accession	ambiguous_bases'):
                col_index = {key: i for i, key in enumerate(each_ref_split)}
            else:
                ref_accession = each_ref_split[0]
                gtdb_taxonomy = each_ref_split[col_index['gtdb_taxonomy']]
                if ref_accession in ref_tree_gnm_set:
                    gnm_p, gnm_c, gnm_o, gnm_f, gnm_g = sep_taxon_str(gtdb_taxonomy)

                    gnm_rooting_rank = ''
                    if rooting_rank == 'p':
                        gnm_rooting_rank = gnm_p
                    elif rooting_rank == 'c':
                        gnm_rooting_rank = gnm_c
                    elif rooting_rank == 'o':
                        gnm_rooting_rank = gnm_o
                    elif rooting_rank == 'f':
                        gnm_rooting_rank = gnm_f
                    elif rooting_rank == 'g':
                        gnm_rooting_rank = gnm_g

                    # rooting_rank
                    if gnm_rooting_rank in rooting_rank_taxon_dict:
                        if gnm_rooting_rank not in counted_taxons_rooting_rank:
                            counted_taxons_rooting_rank.add(gnm_rooting_rank)
                            canditate_gnms_rooting_rank[ref_accession] = gnm_rooting_rank

    ref_tree_rooting_rank = subset_and_rename_tree(gtdb_ref_tree, canditate_gnms_rooting_rank, canditate_gnms_rooting_rank)

    # get the smallest out group taxon set
    smallest_outgroup_taxon_list = get_smallest_outgroup(ref_tree_rooting_rank)

    user_gnm_taxon_dict_rooting_rank = dict()
    if rooting_rank == 'p':
        user_gnm_taxon_dict_rooting_rank = user_gnm_taxon_dict_p
    elif rooting_rank == 'c':
        user_gnm_taxon_dict_rooting_rank = user_gnm_taxon_dict_c
    elif rooting_rank == 'o':
        user_gnm_taxon_dict_rooting_rank = user_gnm_taxon_dict_o
    elif rooting_rank == 'f':
        user_gnm_taxon_dict_rooting_rank = user_gnm_taxon_dict_f
    elif rooting_rank == 'g':
        user_gnm_taxon_dict_rooting_rank = user_gnm_taxon_dict_g

    # get the smallest out group genome set
    out_group_gnm_set_1 = set()
    out_group_gnm_set_2 = set()
    for each_rooting_rank_taxon in user_gnm_taxon_dict_rooting_rank:
        gnm_member_set = user_gnm_taxon_dict_rooting_rank[each_rooting_rank_taxon]
        if each_rooting_rank_taxon in smallest_outgroup_taxon_list:
            out_group_gnm_set_1.update(gnm_member_set)
        else:
            out_group_gnm_set_2.update(gnm_member_set)

    # select the smaller one as outgroup
    if len(out_group_gnm_set_1) < len(out_group_gnm_set_2):
        out_group_gnm_set = out_group_gnm_set_1
    else:
        out_group_gnm_set = out_group_gnm_set_2

    # root user tree with identified out group genomes
    root_with_outgroup(input_unrooted_tree, out_group_gnm_set, add_root_branch, rooted_tree)


if __name__ == '__main__':

    RootTreeGTDB226_parser = argparse.ArgumentParser(usage=RootTreeGTDB226_usage)
    RootTreeGTDB226_parser.add_argument('-tree',        required=True,                         help='input unrooted tree')
    RootTreeGTDB226_parser.add_argument('-tax',         required=False, default='fna',         help='leaf taxon')
    RootTreeGTDB226_parser.add_argument('-db',          required=True,                         help='GTDB database files')
    RootTreeGTDB226_parser.add_argument('-d',           required=False, default=None,          help='domain, either ar or bac')
    RootTreeGTDB226_parser.add_argument('-add_root',    required=False, action='store_true',   help='add the root branch')
    RootTreeGTDB226_parser.add_argument('-o',           required=True,                         help='output folder')
    args = vars(RootTreeGTDB226_parser.parse_args())
    RootTreeGTDB226(args)



import os
import glob
import argparse
from Bio import SeqIO
import multiprocessing as mp
from distutils.spawn import find_executable


ExtractMarkerSeq_usage = '''
============================ ExtractMarkerSeq example commands ============================

Dependencies: blastp

BioSAK ExtractMarkerSeq -m marker_ref_seq -mx fa -aa faa_files -aax faa -o op_dir -e "1e-30" -t 6

===========================================================================================
'''


def check_dependencies(program_list):
    not_detected_programs = []
    for needed_program in program_list:
        if find_executable(needed_program) is None:
            not_detected_programs.append(needed_program)

    if not_detected_programs != []:
        print('%s not found, program exited!' % ','.join(not_detected_programs))
        exit()


def exe_cmds(cmd_list, num_threads):
    print('Running %s commands with %s cores' % (len(cmd_list), num_threads))
    pool = mp.Pool(processes=num_threads)
    pool.map(os.system, cmd_list)
    pool.close()
    pool.join()


def sep_path_basename_ext(file_in):
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'
    file_basename, file_extension = os.path.splitext(file_name)
    return file_path, file_basename, file_extension


def select_seq(seq_file, id_file,select_option, output_file, one_line, in_fastq):

    # get provided id list
    seq_id_list = set()
    for seq_id in open(id_file):
        seq_id_list.add(seq_id.strip())

    seq_in_format = 'fasta'
    if in_fastq is True:
        seq_in_format = 'fastq'

    # extract sequences
    output_file_handle = open(output_file, 'w')
    for seq_record in SeqIO.parse(seq_file, seq_in_format):
        seq_id = seq_record.id
        if select_option == 1:
            if seq_id in seq_id_list:

                if in_fastq is False:
                    if one_line is False:
                        SeqIO.write(seq_record, output_file_handle, 'fasta')
                    else:
                        SeqIO.write(seq_record, output_file_handle, 'fasta-2line')
                else:
                    SeqIO.write(seq_record, output_file_handle, 'fastq')

        if select_option == 0:
            if seq_id not in seq_id_list:

                if in_fastq is False:
                    if one_line is False:
                        SeqIO.write(seq_record, output_file_handle, 'fasta')
                    else:
                        SeqIO.write(seq_record, output_file_handle, 'fasta-2line')
                else:
                    SeqIO.write(seq_record, output_file_handle, 'fastq')
    output_file_handle.close()


def ExtractMarkerSeq(args):

    marker_seq_dir          = args['m']
    marker_seq_ext          = args['mx']
    faa_file_dir            = args['aa']
    faa_file_ext            = args['aax']
    op_dir                  = args['o']
    e_value                 = args['e']
    num_of_threads          = args['t']
    force_overwrite         = args['f']

    # check dependencies
    check_dependencies(['blastp'])

    # get marker id set
    marker_seq_re   = '%s/*.%s' % (marker_seq_dir, marker_seq_ext)
    marker_seq_list = [os.path.basename(file_name) for file_name in glob.glob(marker_seq_re)]
    marker_id_set = set()
    for each_marker_seq_file in marker_seq_list:
        marker_seq_path, marker_seq_basename, marker_seq_ext = sep_path_basename_ext(each_marker_seq_file)
        marker_id_set.add(marker_seq_basename)

    # get gnm id list
    faa_file_re   = '%s/*.%s' % (faa_file_dir, faa_file_ext)
    faa_file_list = [os.path.basename(file_name) for file_name in glob.glob(faa_file_re)]
    gnm_set = set()
    for each_faa_file in faa_file_list:
        faa_path, faa_basename, faa_ext = sep_path_basename_ext(each_faa_file)
        gnm_set.add(faa_basename)
    gnm_id_list_sorted = sorted([i for i in gnm_set])

    # define output dir
    blastp_cmd_txt                       = '%s/blastp_cmds_%s.txt'                  % (op_dir, (len(gnm_id_list_sorted)*len(marker_id_set)))
    pwd_combined_protein                 = '%s/combined.faa'                        % op_dir
    blast_op_dir                         = '%s/s01_blast_op'                        % op_dir
    best_hit_id_by_marker_dir            = '%s/s02_identified_marker_id'            % op_dir
    best_hit_seq_by_marker_dir           = '%s/s03_identified_marker_seq'           % op_dir
    best_hit_seq_by_marker_dir_renamed   = '%s/s04_identified_marker_seq_renamed'   % op_dir

    # create folder
    if force_overwrite is True:
        if os.path.isdir(op_dir) is True:
            os.system('rm -r %s' % op_dir)
        os.system('mkdir %s' % op_dir)
        os.system('mkdir %s' % blast_op_dir)
    else:
        if os.path.isdir(op_dir) is False:
            os.system('mkdir %s' % op_dir)
        if os.path.isdir(blast_op_dir) is False:
            os.system('mkdir %s' % blast_op_dir)

    os.system('cat %s/*.%s > %s' % (faa_file_dir, faa_file_ext, pwd_combined_protein))

    # get blastp command
    blast_cmd_list = []
    blast_op_to_cmd_dict = dict()
    blastp_cmd_txt_handle = open(blastp_cmd_txt, 'w')
    for gnm_id in gnm_id_list_sorted:
        for each_cog in marker_id_set:
            pwd_blast_op = '%s/%s_vs_%s_blastp.txt'                                                 % (blast_op_dir, gnm_id, each_cog)
            blastp_cmd   = 'blastp -subject %s/%s.fa -evalue %s -outfmt 6 -query %s/%s.faa -out %s' % (marker_seq_dir, each_cog, e_value, faa_file_dir, gnm_id, pwd_blast_op)
            blast_op_to_cmd_dict[pwd_blast_op] = blastp_cmd
            blastp_cmd_txt_handle.write(blastp_cmd + '\n')
            blast_cmd_list.append(blastp_cmd)
    blastp_cmd_txt_handle.close()

    # run blastp
    if force_overwrite is True:
        exe_cmds(blast_cmd_list, num_of_threads)
    else:
        cmds_to_rerun = []
        num_of_good_ones = 0
        for each_blast_op in blast_op_to_cmd_dict:

            look_good = False
            if os.path.isfile(each_blast_op) is True:
                look_good = True
                num_of_good_ones += 1

            if look_good is False:
                cmds_to_rerun.append(blast_op_to_cmd_dict[each_blast_op])

        print('Detected blastp outputs: %s' % num_of_good_ones)
        exe_cmds(cmds_to_rerun, num_of_threads)

    # get best_hit_dict_by_marker
    best_hit_to_gnm_dict = dict()
    best_hit_dict_by_marker = dict()
    for gnm_id in gnm_id_list_sorted:
        for each_cog in marker_id_set:
            current_blastp_op = '%s/%s_vs_%s_blastp.txt' % (blast_op_dir, gnm_id, each_cog)
            # get best hit
            if os.path.isfile(current_blastp_op) is True:
                best_hit_gene = ''
                best_hit_score = 0
                for each_line in open(current_blastp_op):
                    each_line_split = each_line.strip().split('\t')
                    query_id = each_line_split[0]
                    bit_score = float(each_line_split[11])
                    if bit_score > best_hit_score:
                        best_hit_score = bit_score
                        best_hit_gene = query_id

                if best_hit_gene != '':
                    best_hit_to_gnm_dict[best_hit_gene] = gnm_id

                    if each_cog not in best_hit_dict_by_marker:
                        best_hit_dict_by_marker[each_cog] = [best_hit_gene]
                    else:
                        best_hit_dict_by_marker[each_cog].append(best_hit_gene)

    # create output dir
    if os.path.isdir(best_hit_id_by_marker_dir) is False:
        os.system('mkdir %s' % best_hit_id_by_marker_dir)
    if os.path.isdir(best_hit_seq_by_marker_dir) is False:
        os.system('mkdir %s' % best_hit_seq_by_marker_dir)
    if os.path.isdir(best_hit_seq_by_marker_dir_renamed) is False:
        os.system('mkdir %s' % best_hit_seq_by_marker_dir_renamed)

    # write out best hits and extract sequences
    processing_index = 1
    for each_marker in best_hit_dict_by_marker:
        print('Extracting marker sequence %s/%s: %s' % (processing_index, len(best_hit_dict_by_marker), each_marker))
        processing_index += 1

        current_m_hit_list      = best_hit_dict_by_marker[each_marker]
        marker_hits_txt         = ('%s/%s.txt' % (best_hit_id_by_marker_dir,          each_marker)).replace(':', '')
        marker_hits_seq         = ('%s/%s.fa'  % (best_hit_seq_by_marker_dir,         each_marker)).replace(':', '')
        marker_hits_seq_renamed = ('%s/%s.fa'  % (best_hit_seq_by_marker_dir_renamed, each_marker)).replace(':', '')

        with open(marker_hits_txt, 'w') as marker_hits_txt_handle:
            marker_hits_txt_handle.write('\n'.join(current_m_hit_list))

        # extract sequences
        select_seq(pwd_combined_protein, marker_hits_txt, 1, marker_hits_seq, True, False)

        # rename sequences
        marker_hits_seq_renamed_handle = open(marker_hits_seq_renamed, 'w')
        for each_seq in SeqIO.parse(marker_hits_seq, 'fasta'):
            seq_id = each_seq.id
            seq_gnm = best_hit_to_gnm_dict[seq_id]
            marker_hits_seq_renamed_handle.write('>%s\n' % seq_gnm)
            marker_hits_seq_renamed_handle.write('%s\n' % str(each_seq.seq))
        marker_hits_seq_renamed_handle.close()

    print('Done!')


if __name__ == '__main__':

    # initialize the options parser
    parser = argparse.ArgumentParser()
    parser.add_argument('-m',               required=True,                          help='marker seq dir')
    parser.add_argument('-mx',              required=True,                          help='marker seq ext')
    parser.add_argument('-aa',              required=True,                          help='faa file dir')
    parser.add_argument('-aax',             required=True,                          help='faa file ext')
    parser.add_argument('-o',               required=True,                          help='output dir')
    parser.add_argument('-e',               required=True,  default=1e-30,          help='e-value cutoff, default: 1e-30')
    parser.add_argument('-t',               required=True,  type=int,               help='num of threads')
    parser.add_argument('-f',               required=False, action="store_true",    help='force overwrite')
    args = vars(parser.parse_args())
    ExtractMarkerSeq(args)


'''

conda activate mypy3env
cd /home-user/wzsong/DateArTree
python3 MarkerRef2Tree.py -m Marker_set_2_Betts_2018_29_arCOG -mx fa -aa /home-user/wzsong/DateArTree/01_genome_selection_Prokka/d__Archaea_o_rs_133_gnms_plus_27_mito_faa_files -aax faa -o Marker_set_2_Betts_2018_29_arCOG_Marker2Tree_e30_demo -e 30 -t 24 -pl /home-user/wzsong/Scripts/catfasta2phyml.pl
submitHPC.sh --cmd "python3 MarkerRef2Tree.py -m Marker_set_2_Betts_2018_29_arCOG -mx fa -aa /home-user/wzsong/DateArTree/01_genome_selection_Prokka/d__Archaea_o_rs_133_gnms_plus_27_mito_faa_files -aax faa -g /home-user/wzsong/DateArTree/gnm_group.txt -o Marker_set_2_Betts_2018_29_arCOG_Marker2Tree_e30_demo -e 30 -t 24 -pl /home-user/wzsong/Scripts/catfasta2phyml.pl" -n 24 -c Marker_set_2_Betts_2018_29_arCOG_Marker2Tree_e30_demo

cd /home-user/wzsong/DateArTree
python3 MarkerRef2Tree.py -m Marker_set_2_Betts_2018_29_arCOG -mx fa -aa /home-user/wzsong/DateArTree/01_genome_selection_Prokka/d__Archaea_o_rs_133_gnms_plus_27_mito_faa_files -aax faa -g /home-user/wzsong/DateArTree/gnm_group.txt -o Marker_set_2_Betts_2018_29_arCOG_Marker2Tree_e30_demo -e 30 -t 12 -pl /home-user/wzsong/Scripts/catfasta2phyml.pl -g gnm_group.txt -skip_align_trim -jst 6 -qsub

cd /Users/songweizhi/Desktop/demo
python3 /Users/songweizhi/PycharmProjects/TreeSAK/TreeSAK/MarkerRef2Tree.py -m Marker_set_2_Betts_2018_29_arCOG -mx fa -aa d__Archaea_o_rs_133_gnms_plus_27_mito_faa_files -aax faa -g gnm_group.txt -o Marker_set_2_Betts_2018_29_arCOG_Marker2Tree_e30_demo -e 30 -t 10 -pl /Users/songweizhi/Scripts/catfasta2phyml.pl -g gnm_group.txt -skip_align_trim -jst 6 -qsub

'''



from __future__ import print_function
import os
import glob
import numpy
import argparse
import subprocess
from ete3 import Tree
import multiprocessing as mp
from operator import itemgetter
from collections import defaultdict


SplitScore2_usage = '''
======================== SplitScore2 example commands ========================

TreeSAK SplitScore2 -i step1_op_dir -g gnm_cluster.tsv -k gnm_taxon.txt -f -t 10 -o step_2_op_dir -c 25,50,75
TreeSAK SplitScore2 -i step1_op_dir -g gnm_cluster.tsv -k gnm_taxon.txt -f -t 10 -o step_2_op_dir -c 20,40,60,80

# format of gnm_cluster.tsv (tab separated)
GCA_013330055.1 c01_UBA8516
GCA_023251795.1 c01_UBA8516
GCA_023251295.1 c01_UBA8516
GCA_005877305.1 c02_TA-20
GCA_013287585.1 c02_TA-20

# gnm_taxon.txt: GTDB format

# install R packages
install.packages("optparse")
install.packages("plyr")
install.packages("dbplyr")
install.packages("dplyr")
install.packages("tidyr")
install.packages("ggplot2")
install.packages("data.table")
install.packages("RColorBrewer")
install.packages("gplots")
install.packages("ape")

# As described in the Undinarchaeota paper (Nina Dombrowski 2020, NC)

=============================================================================
'''


def sep_path_basename_ext(file_in):
    f_path, file_name = os.path.split(file_in)
    if f_path == '':
        f_path = '.'
    f_base, f_ext = os.path.splitext(file_name)
    return f_path, f_base, f_ext


def check_executables(program_list):

    not_detected_programs = []
    for needed_program in program_list:

        if subprocess.call(['which', needed_program], stdout=open(os.devnull, 'wb')) != 0:
            not_detected_programs.append(needed_program)

    if not_detected_programs != []:
        print('%s not detected, program exited!' % ','.join(not_detected_programs))
        exit()


def gtdb_gnm_metadata_parser(gtdb_genome_metadata):

    genome_to_taxon_dict = {}
    genome_to_biosample_dict = {}
    genome_to_completeness_dict = {}
    genome_to_contamination_dict = {}
    col_index = {}
    for each_ref in open(gtdb_genome_metadata):
        each_ref_split = each_ref.strip().split('\t')
        if each_ref.startswith('accession'):
            col_index = {key: i for i, key in enumerate(each_ref_split)}
        else:
            ref_accession = each_ref_split[0][3:]
            gnm_completeness = float(each_ref_split[2])
            gnm_contamination = float(each_ref_split[3])
            gtdb_taxon = each_ref_split[col_index['gtdb_taxonomy']]
            ncbi_biosample = each_ref_split[col_index['ncbi_biosample']]
            genome_to_taxon_dict[ref_accession] = gtdb_taxon
            genome_to_completeness_dict[ref_accession] = gnm_completeness
            genome_to_contamination_dict[ref_accession] = gnm_contamination
            genome_to_biosample_dict[ref_accession] = ncbi_biosample

    return genome_to_completeness_dict, genome_to_contamination_dict, genome_to_taxon_dict, genome_to_biosample_dict


def get_rename_dict(tree_str_in, mag_cluster_dict, gtdb_gnm_tax_dict):

    # rename dict: {'old_name':'new_name'}

    leaf_rename_dict = {}
    for leaf in Tree(tree_str_in, format=1):

        leaf_name_gnm = '_'.join(leaf.name.split('_')[:-1])
        leaf_cluster = mag_cluster_dict.get(leaf_name_gnm, 'cluster_0')

        # get mag_taxon_str
        gnm_taxon_str = 'NA'
        if leaf_name_gnm in gtdb_gnm_tax_dict:
            gnm_taxon_str = gtdb_gnm_tax_dict[leaf_name_gnm]

        # get mag_taxon_str (GCA GCF things)
        if gnm_taxon_str == 'NA':
            mag_id_no_ext_no_source_GCF = leaf_name_gnm.replace('GCA', 'GCF')
            if mag_id_no_ext_no_source_GCF in gtdb_gnm_tax_dict:
                gnm_taxon_str = gtdb_gnm_tax_dict[mag_id_no_ext_no_source_GCF]

        gnm_taxon_str_no_space = gnm_taxon_str.replace(' ', '_')
        gnm_taxon_str_no_space = gnm_taxon_str_no_space.replace(';', '|')
        leaf_name_new = '%s|%s|strain__%s' % (leaf_cluster, gnm_taxon_str_no_space, '_'.join(leaf.name.split('_')[:-1]))

        leaf_rename_dict[leaf.name] = leaf_name_new

    return leaf_rename_dict


def rename_tree(tree_str_in, rename_dict):

    t_in = Tree(tree_str_in, format=1)
    for leaf in t_in:
        leaf_name = leaf.name
        leaf_name_new = rename_dict.get(leaf_name, leaf_name)
        leaf.name = leaf_name_new

    return t_in.write()


def parse_taxonomy(taxon_name):  # given a taxon name, try to return whatever taxonomic info is available as a list starting with the highest level classification and going lower (or a map?)

    name_elements = taxon_name.split('|')
    if (len(name_elements) < 8) or (len(name_elements) > 9):
        print("Nonstandard!")
        quit()

    name_map = dict()
    name_map['cluster'] = name_elements[0]
    name_map['domain']  = name_elements[1]
    name_map['phylum']  = name_elements[2]
    name_map['class']   = name_elements[3]
    name_map['order']   = name_elements[4]
    name_map['family']  = name_elements[5]
    name_map['genus']   = name_elements[6]
    name_map['species'] = name_elements[7]
    if len(name_elements) == 9:
        name_map['ncbi_id'] = name_elements[8]
    return name_map


def summarize_taxonomy(name_list, tax_level, name_to_tax_dict):  # take a list of names from a clade and summarize taxonomic info (labels and their frequencies)
    total_size = len(name_list)  # it perhaps makes sense to normalize by the size of the clade
    breakdown = {}
    for name in name_list:
        info = name_to_tax_dict[name]
        if info[tax_level] in breakdown:
            breakdown[info[tax_level]] += 1.0 / float(total_size)
        else:
            breakdown[info[tax_level]] = 1.0 / float(total_size)
    return breakdown


def count_sister_taxa(target_label, tree_in_ml, tree_in_bs, output_file):

    # edit target_label to make the comparisons at a desired taxonomic level
    # compute the most frequent sister group of each (monophyletic?) group on the tree, to identify trends in gene transfers, "unstable" taxa, etc.

    # read the ML tree, set up the taxonomy stuff, and calculate the number of clades per label, and the sizes of those clades (to report at the end)
    labels = {}
    name_to_tax_info = defaultdict(dict)
    all_tree_leaf_names = []
    ml_tree = Tree(tree_in_ml)  # note that ete3 treats this input tree as rooted
    for leaf in ml_tree:
        taxonomy = parse_taxonomy(leaf.name)
        name_to_tax_info[leaf.name] = taxonomy
        all_tree_leaf_names.append(leaf.name)
        leaf.add_feature("tax", taxonomy[target_label])
        labels[taxonomy[target_label]] = 1
    groups = labels.keys()

    # compute the number of clades (weizhi: monophyletic group) per label in the ML tree, and their sizes
    ML_groups = defaultdict(list)  # the list is the size of each clade, len(list) is the number of clades for that label in the ML tree
    # ML_groups: the number of leaves in each monophyletic groups of the corresponding target_label (e.g. genus)
    for label in groups:
        node_num = 0
        for monophyletic_clade in ml_tree.get_monophyletic(values=[label], target_attr="tax"):  # get monophyletic groups for each target_label (e.g. genus)
            size_clade = 0  # get the number of leaf (size_clade) in the monophyletic group
            for leaf in monophyletic_clade:
                size_clade += 1
            ML_groups[label].append(size_clade)
            node_num += 1

    summary = defaultdict(dict)
    clades_per_group = defaultdict(list)
    treeNum = -1
    for line in open(tree_in_bs):  # read in each bootstrap tree

        treeNum += 1
        tree = Tree(line.rstrip())
        for leaf in tree:
            tax = name_to_tax_info[leaf.name]  # this should set up taxonomy correctly...
            leaf.add_feature("tax", tax[target_label])  # this adds a feature called tax to the leaf, with the attribute of the phylum name
        for label in groups:
            clades_per_group[label].append(0.0)  # setup the clade counting for this particular tree
        tree.unroot()  # Weizhi: why is this

        # iterate over groups that are monophyletic for the taxon label of choice.
        # Choose the smallest sister branch for the comparison. (Assume root is within the larger sister clade (Weizhi:why?))
        for label in groups:
            monophyletic_clade_index = 1
            for monophyletic_clade in tree.get_monophyletic(values=[label], target_attr="tax"):  # node: monophyletic clade
                clades_per_group[label][treeNum] += 1.0
                sister_clades = monophyletic_clade.get_sisters()
                monophyletic_clade_index += 1
                sister_index = 1
                for each_sister in sister_clades:
                    current_sister_leaf_list = []
                    for leaf in each_sister:
                        current_sister_leaf_list.append(leaf.name)
                    sister_index += 1

                if monophyletic_clade.is_root():  # monophyletic clade is root
                    continue

                # Weizhi: bifurcation
                elif len(sister_clades) == 1:  # not at the trifurcation. Do something a bit hacky to find the bigger sister clade

                    taxa_in_sister = []
                    for leaf in sister_clades[0]:
                        taxa_in_sister.append(leaf.name)

                    size_sister = len(taxa_in_sister)

                    taxa_in_group = []
                    for leaf in monophyletic_clade:
                        taxa_in_group.append(leaf.name)

                    taxa_in_other_groups = []  # what does OG mean? (other groups?)
                    for leaf_name in all_tree_leaf_names:
                        if leaf_name in taxa_in_sister:
                            continue
                        elif leaf_name in taxa_in_group:
                            continue
                        else:
                            taxa_in_other_groups.append(leaf_name)
                    size_other_groups = len(taxa_in_other_groups)

                    sister_tax = {}  # taxa in the smaller groups (either the sister group or the OG)
                    if size_other_groups > size_sister:
                        sister_tax = summarize_taxonomy(taxa_in_sister, target_label, name_to_tax_info)
                    else:
                        sister_tax = summarize_taxonomy(taxa_in_other_groups, target_label, name_to_tax_info)

                    # store the tax info of the sister group
                    for element in sister_tax:
                        if element in summary[label]:
                            summary[label][element] += sister_tax[element]
                        else:
                            summary[label][element] = sister_tax[element]

                else:  # trifurcation in tree. Just treat the two sisters in the same way.

                    taxa_in_sisters_1 = []
                    for leaf in sister_clades[0]:
                        taxa_in_sisters_1.append(leaf.name)

                    taxa_in_sisters_2 = []
                    for leaf in sister_clades[1]:
                        taxa_in_sisters_2.append(leaf.name)

                    # get the size of two sisters
                    size_s1 = len(taxa_in_sisters_1)
                    size_s2 = len(taxa_in_sisters_2)

                    # get taxa in the smaller sister group
                    sister_tax = {}
                    if size_s1 > size_s2:
                        sister_tax = summarize_taxonomy(taxa_in_sisters_2, target_label, name_to_tax_info)
                    else:
                        sister_tax = summarize_taxonomy(taxa_in_sisters_1, target_label, name_to_tax_info)

                    for element in sister_tax:
                        if element in summary[label]:
                            summary[label][element] += sister_tax[element]
                        else:
                            summary[label][element] = sister_tax[element]

    # now print out some kind of summary. For each label, the sorted list of sister taxa and their frequencies?
    outh = open(output_file, "w")
    for label in summary:
        num_groups = len(ML_groups[label])
        size_str = ''
        if num_groups == 1:
            size_str = ML_groups[label][0]
        else:
            size_str = ','.join(str(x) for x in (sorted(ML_groups[label], reverse=True)))

        avg_num_clades   = float("{0:.4f}".format(numpy.mean(clades_per_group[label])))
        total_num_clades = numpy.sum(clades_per_group[label])
        sorted_sisters   = sorted(summary[label].items(), key=itemgetter(1), reverse=True)

        for tup in sorted_sisters:
            double_normalize = float(tup[1]) / float(total_num_clades)  # normalize the frequencies by the total number of clades, to account for different bootstrap numbers/MCMC sample numbers
            double_normalize = float("{0:.4f}".format(double_normalize))
            str_to_write     = '%s\t%s\t%s\t%s\t%s\t%s' % (label, tup[0], float("{0:.4f}".format(tup[1])), avg_num_clades, double_normalize, size_str)
            outh.write(str_to_write + '\n')
    outh.close()


def count_sister_taxa_worker(arg_list):

    mag_cluster_dict                = arg_list[0]
    gnm_tax_dict                    = arg_list[1]
    tree_ml                         = arg_list[2]
    ufboot_file                     = arg_list[3]
    target_label                    = arg_list[4]
    tree_ml_renamed                 = arg_list[5]
    ufboot_file_renamed             = arg_list[6]
    count_sister_taxa_op_txt        = arg_list[7]
    gene_id                         = arg_list[8]
    renamed_gnm_to_cluster_dir      = arg_list[9]

    # rename ml tree
    tree_ml_renamed_handle = open(tree_ml_renamed, 'w')
    current_tree_rename_dict = get_rename_dict(tree_ml, mag_cluster_dict, gnm_tax_dict)
    tree_ml_str_renamed = rename_tree(tree_ml, current_tree_rename_dict)
    tree_ml_renamed_handle.write(tree_ml_str_renamed + '\n')
    tree_ml_renamed_handle.close()

    current_renamed_gnm_to_cluster_txt = '%s/%s.txt' % (renamed_gnm_to_cluster_dir, gene_id)
    current_renamed_gnm_to_cluster_txt_handle = open(current_renamed_gnm_to_cluster_txt, 'w')
    for each_leaf in current_tree_rename_dict:
        renamed_leaf = current_tree_rename_dict[each_leaf]
        cluster_id = renamed_leaf.split('|')[0]
        current_renamed_gnm_to_cluster_txt_handle.write('%s\t%s\n' % (renamed_leaf, cluster_id))
    current_renamed_gnm_to_cluster_txt_handle.close()

    # rename ufboot trees
    ufboot_file_renamed_handle = open(ufboot_file_renamed, 'w')
    for each_tree in open(ufboot_file):
        tree_str = each_tree.strip()
        current_tree_rename_dict = get_rename_dict(tree_str, mag_cluster_dict, gnm_tax_dict)
        tree_str_renamed = rename_tree(tree_str, current_tree_rename_dict)
        ufboot_file_renamed_handle.write(tree_str_renamed + '\n')
    ufboot_file_renamed_handle.close()

    # count_sister_taxa
    count_sister_taxa(target_label, tree_ml_renamed, ufboot_file_renamed, count_sister_taxa_op_txt)


def run_count_sister_taxa(gtdb_classification_txt, hog_list, contree_dir, ufboot_dir, gnm_cluster_txt, target_label, num_threads, output_dir, force_overwrite):

    # define file name
    renamed_gnm_to_cluster_dir      = '%s/renamed_genome_to_cluster'            % output_dir
    renamed_gnm_to_cluster_tmp_txt  = '%s/renamed_genome_to_cluster_tmp.txt'    % output_dir
    renamed_gnm_to_cluster_txt      = '%s/renamed_genome_to_cluster.txt'        % output_dir
    renamed_gnm_to_cluster_iTOL_txt = '%s/renamed_genome_to_cluster_iTOL.txt'   % output_dir
    renamed_contree_dir             = '%s/renamed_contree'                      % output_dir
    renamed_ufboot_dir              = '%s/renamed_ufboot'                       % output_dir
    count_sister_taxa_op_dir        = '%s/count_sister_taxa_op'                 % output_dir

    if os.path.isdir(output_dir) is True:
        if force_overwrite is True:
            os.system('rm -r %s' % output_dir)
        else:
            print('%s exist, program exited!' % output_dir)
            exit()

    os.mkdir(output_dir)
    os.mkdir(renamed_contree_dir)
    os.mkdir(renamed_ufboot_dir)
    os.mkdir(count_sister_taxa_op_dir)
    os.mkdir(renamed_gnm_to_cluster_dir)

    ####################################################################################################################

    gnm_tax_dict = {}
    for each in open(gtdb_classification_txt):
        if not each.startswith('user_genome'):
            each_split = each.strip().split('\t')
            gnm_tax_dict[each_split[0]] = each_split[1]

    ####################################################################################################################

    mag_cluster_dict = {}
    for each_gnm in open(gnm_cluster_txt):
        each_gnm_split = each_gnm.strip().split('\t')
        mag_cluster_dict[each_gnm_split[0]] = each_gnm_split[1]

    argument_lol = []
    for og_id in hog_list:

        # define file name
        tree_ml                  = '%s/%s.contree'               % (contree_dir, og_id)
        ufboot_file              = '%s/%s.ufboot'                % (ufboot_dir, og_id)
        tree_ml_renamed          = '%s/%s_renamed.contree'       % (renamed_contree_dir, og_id)
        ufboot_file_renamed      = '%s/%s_renamed.ufboot'        % (renamed_ufboot_dir, og_id)
        count_sister_taxa_op_txt = '%s/%s_count_sister_taxa.txt' % (count_sister_taxa_op_dir, og_id)

        if os.path.isfile(tree_ml) is False:
            print('%s not found!' % tree_ml)
            exit()

        current_arg_list = [mag_cluster_dict, gnm_tax_dict, tree_ml, ufboot_file, target_label, tree_ml_renamed, ufboot_file_renamed, count_sister_taxa_op_txt, og_id, renamed_gnm_to_cluster_dir]
        argument_lol.append(current_arg_list)

    # run with multiprocessing
    pool = mp.Pool(processes=num_threads)
    pool.map(count_sister_taxa_worker, argument_lol)
    pool.close()
    pool.join()

    # combine renamed_gnm_to_cluster files
    os.system('cat %s/*.txt > %s'                                           % (renamed_gnm_to_cluster_dir, renamed_gnm_to_cluster_tmp_txt))
    os.system('cat %s | sort | uniq > %s'                                   % (renamed_gnm_to_cluster_tmp_txt, renamed_gnm_to_cluster_txt))
    BioSAK_iTOL_cmd = 'BioSAK iTOL -ColorRange -lg %s -lt Cluster -o %s'    % (renamed_gnm_to_cluster_txt, renamed_gnm_to_cluster_iTOL_txt)
    os.system(BioSAK_iTOL_cmd)


def get_taxa_count_stats(step_1_op_dir, hog_list_sorted, get_taxa_count_stats_wd, force_overwrite, TaxaCountStats_Rscript):

    # define input files to R script
    combined_contree_file           = '%s/combined.contree'                     % get_taxa_count_stats_wd
    genes_to_remove_txt             = '%s/Genes_to_remove.txt'                  % get_taxa_count_stats_wd
    list_of_trees_txt               = '%s/List_of_trees.txt'                    % get_taxa_count_stats_wd
    mapping_txt                     = '%s/mapping.txt'                          % get_taxa_count_stats_wd
    marker_list_txt                 = '%s/MarkerList.txt'                       % get_taxa_count_stats_wd
    combined_count_sister_taxa_op   = '%s/combined_count_sister_taxa_op.txt'    % get_taxa_count_stats_wd
    TaxaCountStats_op               = '%s/TaxaCountStats_output.txt'            % get_taxa_count_stats_wd

    if os.path.isdir(get_taxa_count_stats_wd) is True:
        if force_overwrite is True:
            os.system('rm -r %s' % get_taxa_count_stats_wd)
        else:
            print('%s exist, program exited!' % get_taxa_count_stats_wd)
            exit()
    os.mkdir(get_taxa_count_stats_wd)

    cluster_to_domain_dict = {}
    marker_list_txt_handle = open(marker_list_txt, 'w')
    marker_list_txt_handle.write('MarkerID\n')
    list_of_trees_txt_handle = open(list_of_trees_txt, 'w')
    combined_contree_file_handle = open(combined_contree_file, 'w')
    combined_count_sister_taxa_op_handle = open(combined_count_sister_taxa_op, 'w')
    combined_count_sister_taxa_op_handle.write('MarkerID\tGroup_of_interest\tSister_taxa\tNormalized_sum_of_occurances\tsplits\tNormalized2_sum_of_occurances\tClusters\n')
    for each_hog in hog_list_sorted:

        # write out to combined_count_sister_taxa_op
        pwd_count_sister_taxa_op_txt = '%s/count_sister_taxa_op/%s_count_sister_taxa.txt' % (step_1_op_dir, each_hog)
        with open(pwd_count_sister_taxa_op_txt) as count_sister_taxa_op_txt_opened:
            for each_line in count_sister_taxa_op_txt_opened:
                combined_count_sister_taxa_op_handle.write('%s\t%s' % (each_hog, each_line))

        # write out to combined_contree_file
        pwd_renamed_contree_file = '%s/renamed_contree/%s_renamed.contree' % (step_1_op_dir, each_hog)
        with open(pwd_renamed_contree_file, 'r') as pwd_renamed_contree_file_opened:
            combined_contree_file_handle.write(pwd_renamed_contree_file_opened.readline())

        # add to cluster_to_domain_dict
        t_in = Tree(pwd_renamed_contree_file, format=1)
        for leaf in t_in:
            leaf_name_split = leaf.name.split('|')
            cluster_to_domain_dict[leaf_name_split[0]] = leaf_name_split[1]

        # write out to marker_list_txt
        marker_list_txt_handle.write(each_hog + '\n')
        list_of_trees_txt_handle.write(each_hog + '\n')

    marker_list_txt_handle.close()
    list_of_trees_txt_handle.close()
    combined_contree_file_handle.close()
    combined_count_sister_taxa_op_handle.close()

    # prepare mapping_txt
    mapping_txt_handle = open(mapping_txt, 'w')
    mapping_txt_handle.write('Cluster\tDomain\n')
    for each_cluster in cluster_to_domain_dict:
        mapping_txt_handle.write('%s\t%s\n' % (each_cluster, cluster_to_domain_dict[each_cluster]))
    mapping_txt_handle.close()

    # prepare genes_to_remove_txt
    genes_to_remove_txt_handle = open(genes_to_remove_txt, 'w')
    genes_to_remove_txt_handle.write('MarkerID\n')
    genes_to_remove_txt_handle.close()

    # run TaxaCountStats.R
    get_TaxaCountStats_cmd = 'Rscript %s -t %s -l %s -g %s -x %s -s %s -r %s -o %s > /dev/null' % (TaxaCountStats_Rscript, combined_contree_file, list_of_trees_txt, mapping_txt, marker_list_txt, combined_count_sister_taxa_op, genes_to_remove_txt, TaxaCountStats_op)
    print('Running: ' + get_TaxaCountStats_cmd)
    os.system(get_TaxaCountStats_cmd)


def group_marker(taxa_counts_tats_op_txt, marker_seq_dir, marker_rank_cutoff_str, op_dir):

    marker_rank_cutoff_list = marker_rank_cutoff_str.split(',')

    marker_score_dict = dict()
    header_index_dict = dict()
    for each_marker in open(taxa_counts_tats_op_txt):
        each_marker_split = each_marker.replace('\n', '').split('\t')
        if each_marker.startswith('MarkerID\t'):
            header_index_dict = {k: v for v, k in enumerate(each_marker_split)}
        else:
            marker_id    = each_marker_split[header_index_dict['MarkerID']]
            marker_score = int(each_marker_split[header_index_dict['RankA_B']])
            marker_score_dict[marker_id] = marker_score

    marker_list_sorted_best_to_wrost = [i[0] for i in sorted(marker_score_dict.items(), key=lambda x: x[1])]
    marker_list_sorted_wrost_to_best = [i[0] for i in sorted(marker_score_dict.items(), key=lambda x: x[1])][::-1]

    for each_cutoff in marker_rank_cutoff_list:

        marker_num_rounded = round(len(marker_list_sorted_wrost_to_best)*float(each_cutoff)/100)
        marker_list_best   = marker_list_sorted_best_to_wrost[:marker_num_rounded]
        marker_list_worst  = marker_list_sorted_wrost_to_best[:marker_num_rounded]
        seq_dir_best       = '%s/best%s'  % (op_dir, each_cutoff)
        seq_dir_worst      = '%s/worst%s' % (op_dir, each_cutoff)

        os.system('mkdir %s' % seq_dir_best)
        os.system('mkdir %s' % seq_dir_worst)

        # get the best markers
        for bm in marker_list_best:
            os.system('cp %s/%s.fa %s/' % (marker_seq_dir, bm, seq_dir_best))

        # get the worst markers
        for wm in marker_list_worst:
            os.system('cp %s/%s.fa %s/' % (marker_seq_dir, wm, seq_dir_worst))


def SplitScore2(args):

    step_1_op_dir               = args['i']
    gnm_group_txt               = args['g']
    gtdb_classification_txt     = args['k']
    force_overwrite             = args['f']
    num_of_threads              = args['t']
    step_2_op_dir               = args['o']
    marker_rank_cutoff_str      = args['c']
    target_label                = 'cluster'
    check_executables(['Rscript'])

    current_file_path           = '/'.join(os.path.realpath(__file__).split('/')[:-1])
    TaxaCountStats_Rscript      = '%s/TaxaCountStats.R'                                     % current_file_path
    qualified_og_seq_dir        = '%s/qualified_OGs'                                        % step_1_op_dir
    contree_file_re             = '%s/*.contree'                                            % qualified_og_seq_dir
    ufboot_file_re              = '%s/*.ufboot'                                             % qualified_og_seq_dir
    count_sister_taxa_op_dir    = '%s/count_sister_taxa_wd'                                 % step_2_op_dir
    get_taxa_count_stats_op_dir = '%s/get_taxa_count_stats_wd'                              % step_2_op_dir
    TaxaCountStats_output_txt   = '%s/get_taxa_count_stats_wd/TaxaCountStats_output.txt'    % step_2_op_dir

    contree_file_set_base = set()
    for each_contree_file in glob.glob(contree_file_re):
        _, f_base, _ = sep_path_basename_ext(each_contree_file)
        contree_file_set_base.add(f_base)

    ufboot_file_set_base = set()
    for each_ufboot_file in glob.glob(ufboot_file_re):
        _, f_base, _ = sep_path_basename_ext(each_ufboot_file)
        ufboot_file_set_base.add(f_base)

    contree_ufboot_shared = set(contree_file_set_base).intersection(ufboot_file_set_base)
    contree_ufboot_shared_sorted = sorted([i for i in contree_ufboot_shared])

    # create output folder
    if os.path.isdir(step_2_op_dir) is True:
        if force_overwrite is True:
            os.system('rm -r %s' % step_2_op_dir)
        else:
            print('%s exist, program exited!' % step_2_op_dir)
            exit()
    os.mkdir(step_2_op_dir)

    print('Counting sister taxa with %s cores' % num_of_threads)
    run_count_sister_taxa(gtdb_classification_txt, contree_ufboot_shared_sorted, qualified_og_seq_dir, qualified_og_seq_dir, gnm_group_txt, target_label, num_of_threads, count_sister_taxa_op_dir, force_overwrite)

    print('Summarising sister taxa')
    get_taxa_count_stats(count_sister_taxa_op_dir, contree_ufboot_shared_sorted, get_taxa_count_stats_op_dir, force_overwrite, TaxaCountStats_Rscript)

    print('Exporting markers by split score')
    group_marker(TaxaCountStats_output_txt, qualified_og_seq_dir, marker_rank_cutoff_str, step_2_op_dir)

    print('Done!')


if __name__ == '__main__':

    SplitScore2_parser = argparse.ArgumentParser()
    SplitScore2_parser.add_argument('-i', required=True,                        help='output dir from SplitScore1')
    SplitScore2_parser.add_argument('-g', required=True,                        help='genome group')
    SplitScore2_parser.add_argument('-k', required=True,                        help='genome taxon, GTDB format')
    SplitScore2_parser.add_argument('-f', required=False, action="store_true",  help='force overwrite')
    SplitScore2_parser.add_argument('-t', required=False, type=int, default=1,  help='num of threads, default: 1')
    SplitScore2_parser.add_argument('-c', required=False, default='25,50,75',   help='marker ranking cutoffs, default: 25,50,75')
    SplitScore2_parser.add_argument('-o', required=True,                        help='output directory')
    args = vars(SplitScore2_parser.parse_args())
    SplitScore2(args)



import os
import glob
import argparse
from Bio import SeqIO
from Bio import AlignIO


AssessMarkerDeltaLL_usage = '''
============================= AssessMarkerDeltaLL example commands =============================

Dependencies: iqtree

# example commands
TreeSAK AssessMarkerDeltaLL -deltall DeltaLL_stdout.txt -o op_dir -c 25-50-75-100 -mmn 20 -aln trimmed_aln_dir -jst 6 -qsub -f 

================================================================================================
'''


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def catfasta2phy(msa_dir, msa_ext, concatenated_msa_phy, partition_file):

    concatenated_msa_fasta = '%s.fasta' % concatenated_msa_phy
    msa_file_re            = '%s/*.%s'  % (msa_dir, msa_ext)
    msa_file_list          = [os.path.basename(file_name) for file_name in glob.glob(msa_file_re)]
    msa_file_list_sorted   = sorted(msa_file_list)

    complete_gnm_set = set()
    for each_msa_file in msa_file_list:
        pwd_msa = '%s/%s' % (msa_dir, each_msa_file)
        for each_seq in SeqIO.parse(pwd_msa, 'fasta'):
            complete_gnm_set.add(each_seq.id)

    complete_gnm_list_sorted = sorted([i for i in complete_gnm_set])

    # initialize concatenated msa dict
    gnm_to_seq_dict = {i: '' for i in complete_gnm_list_sorted}
    msa_len_dict = dict()
    for each_msa_file in msa_file_list_sorted:
        gene_id = each_msa_file.split('.' + msa_ext)[0]

        # read in msa
        current_msa_len = 0
        current_msa_len_set = set()
        pwd_current_msa = '%s/%s' % (msa_dir, each_msa_file)
        current_msa_seq_dict = dict()
        for each_seq in SeqIO.parse(pwd_current_msa, 'fasta'):
            complete_gnm_set.add(each_seq.id)
            current_msa_seq_dict[each_seq.id] = str(each_seq.seq)
            current_msa_len_set.add(len(each_seq.seq))
            current_msa_len = len(each_seq.seq)

        if len(current_msa_len_set) != 1:
            print('Sequences with different length were found in %s, program exited!' % each_msa_file)
            exit()

        msa_len_dict[gene_id] = current_msa_len

        # add sequence to concatenated msa dict
        for each_gnm in complete_gnm_list_sorted:
            msa_seq = current_msa_seq_dict.get(each_gnm, current_msa_len*'-')
            gnm_to_seq_dict[each_gnm] += msa_seq

    # write out concatenated msa
    concatenated_msa_handle = open(concatenated_msa_fasta, 'w')
    for each_gnm in complete_gnm_list_sorted:
        concatenated_msa_handle.write('>%s\n' % each_gnm)
        concatenated_msa_handle.write('%s\n' % gnm_to_seq_dict[each_gnm])
    concatenated_msa_handle.close()

    # write out partition file
    end_pos = 0
    partition_file_handle = open(partition_file, 'w')
    for each_m in msa_file_list_sorted:
        gene_id = each_m.split('.' + msa_ext)[0]
        current_m_len = msa_len_dict[gene_id]
        partition_file_handle.write('%s = %s-%s\n' % (each_m, (end_pos + 1), (end_pos + current_m_len)))
        end_pos += current_m_len
    partition_file_handle.close()

    # convert msa in fasta to phy
    AlignIO.convert(concatenated_msa_fasta, 'fasta', concatenated_msa_phy, 'phylip-relaxed')


def submit_js(js):
    current_wd = os.getcwd()
    js_path, js_basename, js_ext = sep_path_basename_ext(js)
    os.chdir(js_path)
    os.system('qsub %s%s' % (js_basename, js_ext))
    os.chdir(current_wd)


def AssessMarkerDeltaLL(args):

    deltall_stdout_txt      = args['deltall']
    op_dir                  = args['o']
    deltall_keep_pct_str    = args['c']
    min_marker_num          = args['mmn']
    trimmed_aln_dir         = args['aln']
    js_cpu_num              = args['jst']
    submit_job              = args['qsub']
    force_overwrite         = args['f']

    deltall_keep_pct_list = [int(i) for i in deltall_keep_pct_str.split('-')]

    deltall_stdout_path, deltall_stdout_basename, deltall_stdout_ext = sep_path_basename_ext(deltall_stdout_txt)

    # create dir
    if os.path.isdir(op_dir) is True:
        if force_overwrite is True:
            os.system('rm -r %s' % op_dir)
        else:
            print('output folder detected, program exited!')
            exit()
    os.system('mkdir %s' % op_dir)

    # define file name
    deltall_stdout_summary_txt = '%s/%s_summary.txt' % (op_dir, deltall_stdout_basename)

    # read in deltall_stdout_txt
    deltall_op_dict = dict()
    for each_line in open(deltall_stdout_txt):
        if not ((each_line.startswith('WARNING:')) or (each_line.startswith('awk:'))):
            each_line_split = each_line.strip().split('\t')
            marker_id = each_line_split[0]
            value = float(each_line_split[1])
            if marker_id not in deltall_op_dict:
                deltall_op_dict[marker_id] = [value]
            else:
                deltall_op_dict[marker_id].append(value)

    # assigned score to marker
    metric_1_dict = dict()
    metric_2_dict = dict()
    for each_marker in deltall_op_dict:
        metric_1_value = float("{0:.2f}".format(deltall_op_dict[each_marker][0]))
        metric_2_value = float("{0:.2f}".format(deltall_op_dict[each_marker][1]))
        metric_1_dict[each_marker] = metric_1_value
        metric_2_dict[each_marker] = metric_2_value

    metric_1_dict_sorted = {k: v for k, v in sorted(metric_1_dict.items(), key=lambda item: item[1])[::-1]}
    metric_2_dict_sorted = {k: v for k, v in sorted(metric_2_dict.items(), key=lambda item: item[1])}

    metric_1_score_dict = dict()
    metric_1_score = 1
    for each_marker_1 in metric_1_dict_sorted:
        metric_1_score_dict[each_marker_1] = metric_1_score
        metric_1_score += 1

    metric_2_score_dict = dict()
    metric_2_score = 1
    for each_marker_2 in metric_2_dict_sorted:
        metric_2_score_dict[each_marker_2] = metric_2_score
        metric_2_score += 1

    overall_score_dict = dict()
    for each_marker in deltall_op_dict:
        metric_score_1 = metric_1_score_dict[each_marker]
        metric_score_2 = metric_2_score_dict[each_marker]
        metric_score_overall = metric_score_1 + metric_score_2
        overall_score_dict[each_marker] = metric_score_overall

    overall_score_dict_sorted     = {k: v for k, v in sorted(overall_score_dict.items(), key=lambda item: item[1])}
    marker_list_sorted_by_deltall = [k for k, v in sorted(overall_score_dict.items(), key=lambda item: item[1])]

    # write out summary_txt
    summary_txt_handle = open(deltall_stdout_summary_txt, 'w')
    summary_txt_handle.write('Marker\tmetric1\tmetric1_score\tmetric2\tmetric2_score\toverall_score\n')
    for each_marker in overall_score_dict_sorted:
        metric_value_1 = metric_1_dict[each_marker]
        metric_value_2 = metric_2_dict[each_marker]
        metric_score_1 = metric_1_score_dict[each_marker]
        metric_score_2 = metric_2_score_dict[each_marker]
        metric_score_overall = overall_score_dict_sorted[each_marker]
        summary_txt_handle.write('%s\t%s\t%s\t%s\t%s\t%s\n' % (each_marker, metric_value_1, metric_score_1, metric_value_2, metric_score_2, metric_score_overall))
    summary_txt_handle.close()

    # get qualified marker list
    for each_keep_pct in deltall_keep_pct_list:
        marker_num_to_keep = round(len(marker_list_sorted_by_deltall)*each_keep_pct/100)
        markers_to_keep_id_list = marker_list_sorted_by_deltall[:marker_num_to_keep]

        if marker_num_to_keep < min_marker_num:
            print('Ignored DeltaLL cutoff at %s , the number of qualified markers (%s) less than %s' % (each_keep_pct, marker_num_to_keep, min_marker_num))
        else:
            pwd_aln_dir                     = '%s/%s_DeltaLL_%s_trimmed_aln'                % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)
            pwd_aln_concatenated            = '%s/%s_DeltaLL_%s_concatenated.phy'           % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)
            pwd_aln_concatenated_partitions = '%s/%s_DeltaLL_%s_concatenated_partition.txt' % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)
            pwd_iqtree_guide_tree_wd        = '%s/%s_DeltaLL_%s_iqtree_guide_tree'          % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)
            pwd_iqtree_c60_pmsf_wd          = '%s/%s_DeltaLL_%s_iqtree_C60_PMSF'            % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)
            pwd_iqtree_js                   = '%s/js_%s_DeltaLL_%s_iqtree.sh'               % (op_dir, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)

            # create dir
            os.mkdir(pwd_aln_dir)

            # copy msa of qualified markers
            for each_marker in markers_to_keep_id_list:
                pwd_marker_aln = '%s/%s.aln' % (trimmed_aln_dir, each_marker)
                cp_cmd = 'cp %s %s/' % (pwd_marker_aln, pwd_aln_dir)
                os.system(cp_cmd)

            # concatenate msa
            catfasta2phy(pwd_aln_dir, 'aln', pwd_aln_concatenated, pwd_aln_concatenated_partitions)

            # create dir
            os.mkdir(pwd_iqtree_guide_tree_wd)
            os.mkdir(pwd_iqtree_c60_pmsf_wd)

            # run iqtree
            get_guide_tree_cmd = 'iqtree -s ../%s_DeltaLL_%s_concatenated.phy --prefix guide_tree --seqtype AA -m LG -T %s -B 1000 --alrt 1000 --quiet'                                                                      % (deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct, js_cpu_num)
            get_c60_tree_cmd   = 'iqtree -s ../%s_DeltaLL_%s_concatenated.phy --prefix concatenated --seqtype AA -m LG+G+F+C60 -T %s -B 1000 --alrt 1000 --quiet -ft ../%s_DeltaLL_%s_iqtree_guide_tree/guide_tree.treefile' % (deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct, js_cpu_num, deltall_stdout_basename.split('_DeltaLL_stdout')[0], each_keep_pct)

            # write job script
            with open(pwd_iqtree_js, 'w') as pwd_iqtree_js_handle:
                pwd_iqtree_js_handle.write('#!/bin/bash\n#SBATCH --ntasks 1\n#SBATCH --cpus-per-task %s\n\n' % js_cpu_num)
                pwd_iqtree_js_handle.write('cd %s/%s\n' % (os.getcwd(), pwd_iqtree_guide_tree_wd))
                pwd_iqtree_js_handle.write(get_guide_tree_cmd + '\n\n')
                pwd_iqtree_js_handle.write('cd %s/%s\n' % (os.getcwd(), pwd_iqtree_c60_pmsf_wd))
                pwd_iqtree_js_handle.write(get_c60_tree_cmd + '\n')

            if submit_job is True:
                print(pwd_iqtree_js)
                submit_js(pwd_iqtree_js)
            else:
                print('Job script for running iqtree exported to %s' % pwd_iqtree_js)

    # prepare files for performing dating
    print('Preparing files for performing dating')



if __name__ == '__main__':

    # initialize the options parser
    parser = argparse.ArgumentParser()
    parser.add_argument('-deltall', required=True,                          help='DeltaLL stdout')
    parser.add_argument('-o',       required=True,                          help='output dir')
    parser.add_argument('-c',       required=False, default='25-50-75-100', help='cutoffs, default: 25-50-75-100')
    parser.add_argument('-mmn',     required=False, default=20, type=int,   help='minimal marker number, default: 20')
    parser.add_argument('-aln',     required=True,                          help='faa file dir')
    parser.add_argument('-jst',     required=False, default='6',            help='threads to request in job script, for running iqtree')
    parser.add_argument('-qsub',    required=False, action="store_true",    help='submit job scripts')
    parser.add_argument('-f',       required=False, action="store_true",    help='force overwrite')
    args = vars(parser.parse_args())
    AssessMarkerDeltaLL(args)



import os
import glob
import argparse
import pandas as pd
import plotly.express as px


PlotMcmcNode_usage = '''
========================== PlotMcmcNode example commands ==========================

TreeSAK PlotMcmcNode -i McmcTree_op_files -n topo123_nodes.txt -o topo123_age.pdf

# txt file format (-i, tab separated, no header)
# column 1: mcmc file
# column 2: node id
# column 3: mcmc file description
# column 4: node description

path/to/topo1_clock3_mcmc.txt	t_n171	Topo1	Symbiosis_event_1
path/to/topo1_clock3_mcmc.txt	t_n151	Topo1	Symbiosis_event_2
path/to/topo1_clock3_mcmc.txt	t_n131	Topo1	Symbiosis_event_3
path/to/topo2_clock3_mcmc.txt	t_n171	Topo2	Symbiosis_event_1
path/to/topo2_clock3_mcm.txtc	t_n171	Topo3	Symbiosis_event_1

===================================================================================
'''


def sep_path_basename_ext(file_in):

    # separate path and file name
    f_path, file_name = os.path.split(file_in)
    if f_path == '':
        f_path = '.'

    # separate file basename and extension
    f_base, f_ext = os.path.splitext(file_name)

    return f_path, f_base, f_ext


def plot_distribution(df_txt, output_plot):

    df = pd.read_table(df_txt, sep=',')
    run_id_list = df['Setting'].unique()
    node_id_list = df['Node'].unique()

    # sort dataframe by run id
    df = df.sort_values(by='Setting', ascending=False)

    plot_width  = 900
    plot_height = len(run_id_list)*100
    if plot_height < 360:
        plot_height = 360

    fig = px.violin(df, x="Value", y="Setting", color="Node", points=False, orientation="h", width=plot_width, height=plot_height).update_layout(xaxis_title="Age", yaxis_title="Settings")
    if len(node_id_list) == 1:
        fig.update_traces(side="positive", fillcolor='lightblue', width=1.6, opacity=0.75)
    else:
        fig.update_traces(side="positive", fillcolor='rgba(0,0,0,0)', width=1.6)

    fig.update_traces(showlegend=True)
    fig.layout.template = "simple_white"
    # fig.layout.width = 700
    # fig.layout.height = 750
    # fig.update_xaxes(range=[40, 0])
    # fig.update_layout(margin_t=10, title_text='Demo', title_x=0.5)
    fig.write_image(output_plot)


def PlotMcmcNode(args):

    node_txt    = args['i']
    output_plot = args['o']

    if output_plot[-4:] not in ['.PDF', '.pdf']:
        output_plot = output_plot + '.pdf'
    op_df_tmp = output_plot + '.txt'

    # read in txt file
    file_to_node_dict = dict()
    file_rename_dict = dict()
    node_rename_dict = dict()
    not_found_file_set = set()
    for each_node in open(node_txt):
        line_split = each_node.strip().split()
        mcmc_file           = line_split[0]
        node_id             = line_split[1]

        if os.path.isfile(mcmc_file) is False:
            not_found_file_set.add(mcmc_file)

        _, mcmc_f_base, _       = sep_path_basename_ext(mcmc_file)
        node_id_with_file_name  = '%s_____%s' % (mcmc_file, node_id)

        # get file_to_node_dict
        if mcmc_file not in file_to_node_dict:
            file_to_node_dict[mcmc_file] = set()
        file_to_node_dict[mcmc_file].add(node_id)

        # get file_rename_dict
        if len(line_split) >= 3:
            file_rename_on_plot = line_split[2]
            file_rename_dict[mcmc_file] = file_rename_on_plot

        # get node_rename_dict
        if len(line_split) == 4:
            node_rename_on_plot = line_split[3]
            node_rename_dict[node_id_with_file_name] = node_rename_on_plot

    if len(not_found_file_set) > 0:
        print('The following files not found program exited!')
        print('\n'.join(sorted(list(not_found_file_set))))
        exit()

    found_matched_node = False
    op_df_tmp_handle = open(op_df_tmp, 'w')
    op_df_tmp_handle.write('Value,Node,Setting\n')
    for mcmc_file in file_to_node_dict:
        current_node_set = file_to_node_dict.get(mcmc_file, set())
        file_name_to_plot = file_rename_dict.get(mcmc_file, mcmc_file)
        mcmc_df = pd.read_table(mcmc_file, index_col=0)
        for each_col in mcmc_df:
            if each_col in current_node_set:
                node_desc_to_plot = node_rename_dict.get(('%s_____%s' % (mcmc_file, each_col)), each_col)
                found_matched_node = True
                value_list = mcmc_df[each_col].values
                for each_value in value_list:
                    op_df_tmp_handle.write('%s,%s,%s\n' % (each_value, node_desc_to_plot, file_name_to_plot))
    op_df_tmp_handle.close()

    if found_matched_node is False:
        print('Provided node(s) not found, program exited!')
        exit()

    # plot distribution
    plot_distribution(op_df_tmp, output_plot)

    # remove tmp files
    os.system('rm %s' % op_df_tmp)

    # final report
    print('Plot exported to %s, done!' % output_plot)


if __name__ == '__main__':

    PlotMcmcNode_parser = argparse.ArgumentParser()
    PlotMcmcNode_parser.add_argument('-i',  required=True,                  help='input txt file')
    PlotMcmcNode_parser.add_argument('-o',  required=True,                  help='output plot')
    args = vars(PlotMcmcNode_parser.parse_args())
    PlotMcmcNode(args)



import os
import argparse
from BioSAK.BioSAK_config import config_dict


label_tree_usage = '''
======================== label_tree example commands ========================

module load R

# label tree with customized grouping file
BioSAK label_tree -tree NorthSea.tree -label labels.txt 

# label tree by taxonomic classification at phylum and class levels
BioSAK label_tree -tree NorthSea.tree -taxon GTDB_output.tsv -rank p
BioSAK label_tree -tree NorthSea.tree -taxon GTDB_output.tsv -rank c

# label file format:
label_A,tree_leaf_1
label_B,tree_leaf_2
label_B,tree_leaf_3
label_C,tree_leaf_4

=============================================================================
'''


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def label_tree(args, config_dict):

    tree_in      = args['tree']
    label_file   = args['label']
    leaf_taxon   = args['taxon']
    taxon_rank   = args['rank']
    label_tree_R = config_dict['label_tree_R']

    if (label_file is not None) and (leaf_taxon is None) and (taxon_rank is None):
        label_tree_cmd = 'Rscript %s -t %s -g %s' % (label_tree_R, tree_in, label_file)
        os.system(label_tree_cmd)

    elif (label_file is None) and (leaf_taxon is not None) and (taxon_rank is not None):

        # define tmp file name
        tree_file_path, tree_file_basename, tree_file_extension = sep_path_basename_ext(tree_in)
        taxon_grouping = '%s/%s_%s.txt' % (tree_file_path, tree_file_basename, taxon_rank)

        # read GTDB output into dict
        taxon_assignment_dict = {}
        for each_genome in open(leaf_taxon):
            if not each_genome.startswith('user_genome'):
                each_split = each_genome.strip().split('\t')
                bin_name = each_split[0]

                assignment_full = []
                if len(each_split) == 1:
                    assignment_full = ['d__', 'p__', 'c__', 'o__', 'f__', 'g__', 's__']
                elif (len(each_split) > 1) and (';' in each_split[1]):
                    assignment = each_split[1].split(';')
                    if len(assignment) == 7:
                        assignment_full = assignment
                    if len(assignment) == 6:
                        assignment_full = assignment + ['s__']
                    if len(assignment) == 5:
                        assignment_full = assignment + ['g__', 's__']
                    if len(assignment) == 4:
                        assignment_full = assignment + ['f__', 'g__', 's__']
                    if len(assignment) == 3:
                        assignment_full = assignment + ['o__', 'f__', 'g__', 's__']
                    if len(assignment) == 2:
                        assignment_full = assignment + ['c__', 'o__', 'f__', 'g__', 's__']

                elif (len(each_split) > 1) and (';' not in each_split[1]):
                    assignment_full = [each_split[1]] + ['p__', 'c__', 'o__', 'f__', 'g__', 's__']

                # store in dict
                taxon_assignment_dict[bin_name] = assignment_full

        # get all identified taxon at defined ranks
        rank_to_position_dict = {'d': 0, 'p': 1, 'c': 2, 'o': 3, 'f': 4, 'g': 5, 's': 6}
        specified_rank_pos = rank_to_position_dict[taxon_rank]

        taxon_grouping_handle = open(taxon_grouping, 'w')
        for each_TaxonAssign in taxon_assignment_dict:
            specified_rank_id = taxon_assignment_dict[each_TaxonAssign][specified_rank_pos]
            taxon_grouping_handle.write('%s,%s\n' % (specified_rank_id, each_TaxonAssign))
        taxon_grouping_handle.close()

        # run R script
        label_tree_cmd = 'Rscript %s -t %s -g %s' % (label_tree_R, tree_in, taxon_grouping)
        os.system(label_tree_cmd)

    else:
        print('Please provide either a customized label file or the taxonomy info of tree leaves together with a taxonomic rank')
        print('Program exited!')
        exit()


if __name__ == '__main__':

    # initialize the options parser
    parser = argparse.ArgumentParser(usage=label_tree_usage)

    parser.add_argument('-tree',      required=True,                 help='tree file in newick format')
    parser.add_argument('-label',     required=False,  default=None, help='label file (label,leaf)')
    parser.add_argument('-taxon',     required=False,  default=None, help='taxonomic classification')
    parser.add_argument('-rank',      required=False,  default=None, help='taxonomic rank to label')

    args = vars(parser.parse_args())
    label_tree(args, config_dict)



#!/usr/bin/env python3

import os
import glob
import argparse
from Bio import SeqIO
from time import sleep
import multiprocessing as mp
from datetime import datetime
from BioSAK.global_functions import time_format
from BioSAK.global_functions import force_create_folder
from BioSAK.global_functions import sep_path_basename_ext
from BioSAK.global_functions import get_gene_list_TotalDepth
from BioSAK.global_functions import AnnotateNorm


KEGG_parser_usage = '''
======================================== KEGG example commands =======================================

# Dependencies
module load blast+
module load diamond

# annotation with NCBI blastp (default, for small dataset)
BioSAK KEGG -db_dir path/to/your/KEGG_db_dir -t 6 -seq_in input.faa -depth input.depth

# annotation with Diamond blastp (for big dataset)
BioSAK KEGG -db_dir path/to/your/KEGG_db_dir -t 12 -seq_in faa_folder -x faa -depth depth_files -diamond

# get summary for BlastKOALA/GhostKOALA produced results
BioSAK KEGG -db_dir path/to/your/KEGG_db_dir -t 9 -ko_in user_ko.txt
BioSAK KEGG -db_dir path/to/your/KEGG_db_dir -t 9 -ko_in user_ko_folder -x txt

# Prepare DB files, you need to have the following three files in your KEGG_db_dir:
1. Sequence file, only needed for "-seq_in" mode, DECOMPRESS and RENAME it to kegg_db_seq.fasta
   e.g. prokaryotes.pep.gz (https://www.kegg.jp/kegg/download/Readme/README.fasta)
2. seq2ko file, only needed for "-seq_in" mode, DECOMPRESS and RENAME it to kegg_db_seq2ko.txt
   e.g. prokaryotes.dat.gz (https://www.kegg.jp/kegg/download/Readme/README.fasta)
3. ko00001.keg
   https://www.genome.jp/kegg-bin/download_htext?htext=ko00001&format=htext&filedir=

# How it works:
1. KEGG module uses Blast+/Diamond to get the best hits of query genes in the database with user defined e-value cutoff (default 0.001).
2. The TotalDepth of a KO is calculated by summing up the depth of all genes assigned to it.
3. The percentage of GeneNumber/TotalDepth of genes assigned to a KO is calculated by dividing them 
   by the total number/depth of genes with KO assignment (default) or by all genes in a genome ("-pct_by_all"). 

# Note!!!
1. If you run KEGG annotation for multiple files in a batch manner and want to have their depth info incorporated into the results, 
   you need to provide a folder containing individual depth files for each of your input sequence file.
   Name of the depth file needs to be exactly the same as its corresponding sequence file, except the extension which is ".depth".
2. Diamond requires quite a lot of memory for sequence comparison, especially for huge db file (e.g. KEGG db).
   Remember to request sufficient memory (e.g. 90 or 120gb) in your job script and specify a small number (e.g. -t 6) 
   of jobs executing in parallel. Otherwise, you may see some of your query genomes with no gene been annotated.

# Depth file format (one gene per line, tab separated)
gene_1	30
gene_2	10.58

# To do:
1. level C stats: separate stats for Pathway, Brite and the rests

======================================================================================================
'''


def keep_blast_hit_with_highest_bit_score(file_in, file_out):
    file_out_handle = open(file_out, 'w')
    best_hit_line = ''
    best_hit_query_id = ''
    best_hit_score = 0
    for blast_hit in open(file_in):
        blast_hit_split = blast_hit.strip().split('\t')
        query_id = blast_hit_split[0]
        bit_score = float(blast_hit_split[11])

        if best_hit_query_id == '':
            best_hit_query_id = query_id
            best_hit_line = blast_hit
            best_hit_score = bit_score

        elif (query_id == best_hit_query_id) and (bit_score > best_hit_score):
            best_hit_score = bit_score
            best_hit_line = blast_hit

        elif query_id != best_hit_query_id:
            file_out_handle.write(best_hit_line)
            best_hit_query_id = query_id
            best_hit_line = blast_hit
            best_hit_score = bit_score

    file_out_handle.write(best_hit_line)
    file_out_handle.close()


def run_blast_worker(argument_list):

    pwd_input_file =        argument_list[0]
    run_blast =             argument_list[1]
    run_diamond =           argument_list[2]
    KEGG_DB_seq =           argument_list[3]
    KEGG_DB_seq_diamond =   argument_list[4]
    op_dir =                argument_list[5]
    evalue_cutoff =         argument_list[6]
    threads_num =           argument_list[7]

    ################################################### define file name ###################################################

    input_file_path, in_file_basename, input_file_ext = sep_path_basename_ext(pwd_input_file)

    blast_results =           '%s/%s_KEGG_wd/%s_blast.tab'              % (op_dir, in_file_basename, in_file_basename)
    blast_results_best_hit =  '%s/%s_KEGG_wd/%s_blast_best_hits.tab'    % (op_dir, in_file_basename, in_file_basename)

    # create output folder
    force_create_folder('%s/%s_KEGG_wd' % (op_dir, in_file_basename))

    ########################################## blast against KEGG database (Shan) ##########################################

    if run_blast is True:

        if run_diamond is False:
            blastp_cmd = 'blastp -query %s -db %s -out %s -outfmt 6 -evalue %s -num_alignments 10 -num_threads %s' % (pwd_input_file, KEGG_DB_seq, blast_results, evalue_cutoff, threads_num)
            os.system(blastp_cmd)

        else:
            diamond_cmd = 'diamond blastp -q %s --db %s --out %s --outfmt 6 --evalue %s --block-size 1 --threads %s --quiet' % (pwd_input_file, KEGG_DB_seq_diamond, blast_results, evalue_cutoff, threads_num)
            os.system(diamond_cmd)

        # only keep the best hit
        keep_blast_hit_with_highest_bit_score(blast_results, blast_results_best_hit)


def write_out_stats_GeneNumber(identified_ko_list, ko_to_gene_member_dict, ko_description_dict, stats_file_GeneNumber):

    stats_file_GeneNumber_handle = open(stats_file_GeneNumber, 'w')
    stats_file_GeneNumber_handle.write('KO\tGeneNumber\tDescription\n')
    for ko in identified_ko_list:
        ko_GeneNumber = len(ko_to_gene_member_dict[ko])
        stats_file_GeneNumber_handle.write('%s\t%s\t%s\n' % (ko[2:], ko_GeneNumber, ko_description_dict[ko[2:]]))
    stats_file_GeneNumber_handle.close()


def write_out_stats_TotalDepth(identified_ko_list, ko_to_gene_member_dict, gene_depth_dict, ko_description_dict, stats_file_TotalDepth):

    stats_file_TotalDepth_handle = open(stats_file_TotalDepth, 'w')
    stats_file_TotalDepth_handle.write('KO\tTotalDepth\tDescription\n')
    for ko in identified_ko_list:
        ko_gene_total_depth = 0
        for each_gene in ko_to_gene_member_dict[ko]:
            each_gene_depth = gene_depth_dict[each_gene]
            ko_gene_total_depth += each_gene_depth
        ko_TotalDepth = float("{0:.2f}".format(ko_gene_total_depth))
        stats_file_TotalDepth_handle.write('%s\t%s\t%s\n' % (ko[2:], ko_TotalDepth, ko_description_dict[ko[2:]]))
    stats_file_TotalDepth_handle.close()


def parse_blast_op_worker(argument_list):

    pwd_input_file =        argument_list[0]
    run_blast =             argument_list[1]
    As_description_dict =   argument_list[2]
    Bs_description_dict =   argument_list[3]
    Cs_description_dict =   argument_list[4]
    Ds_description_dict =   argument_list[5]
    D2ABCD_dict =           argument_list[6]
    db_seq_to_KO_dict =     argument_list[7]
    op_dir =                argument_list[8]
    depth_file =            argument_list[9]
    pct_by_all =            argument_list[10]

    ################################################### define file name ###################################################

    input_file_path, in_file_basename, input_file_ext = sep_path_basename_ext(pwd_input_file)

    blast_results_best_hit =             '%s/%s_KEGG_wd/%s_blast_best_hits.tab'                     % (op_dir, in_file_basename, in_file_basename)
    KO_assignment_file_D =               '%s/%s_KEGG_wd/%s_KO_assignment_D.txt'                     % (op_dir, in_file_basename, in_file_basename)
    KO_assignment_file_DCBA =            '%s/%s_KEGG_wd/%s_ko_assignment_ABCD.txt'                  % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_GeneNumber =            '%s/%s_KEGG_wd/%s_ko_stats_A_GeneNumber.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_GeneNumber =            '%s/%s_KEGG_wd/%s_ko_stats_B_GeneNumber.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_GeneNumber =            '%s/%s_KEGG_wd/%s_ko_stats_C_GeneNumber.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_GeneNumber =            '%s/%s_KEGG_wd/%s_ko_stats_D_GeneNumber.txt'               % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_TotalDepth =            '%s/%s_KEGG_wd/%s_ko_stats_A_TotalDepth.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_TotalDepth =            '%s/%s_KEGG_wd/%s_ko_stats_B_TotalDepth.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_TotalDepth =            '%s/%s_KEGG_wd/%s_ko_stats_C_TotalDepth.txt'               % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_TotalDepth =            '%s/%s_KEGG_wd/%s_ko_stats_D_TotalDepth.txt'               % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_GeneNumber_pct =        '%s/%s_KEGG_wd/%s_ko_stats_A_GeneNumber_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_GeneNumber_pct =        '%s/%s_KEGG_wd/%s_ko_stats_B_GeneNumber_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_GeneNumber_pct =        '%s/%s_KEGG_wd/%s_ko_stats_C_GeneNumber_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_GeneNumber_pct =        '%s/%s_KEGG_wd/%s_ko_stats_D_GeneNumber_pct.txt'           % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_TotalDepth_pct =        '%s/%s_KEGG_wd/%s_ko_stats_A_TotalDepth_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_TotalDepth_pct =        '%s/%s_KEGG_wd/%s_ko_stats_B_TotalDepth_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_TotalDepth_pct =        '%s/%s_KEGG_wd/%s_ko_stats_C_TotalDepth_pct.txt'           % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_TotalDepth_pct =        '%s/%s_KEGG_wd/%s_ko_stats_D_TotalDepth_pct.txt'           % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_GeneNumber_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_A_GeneNumber_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_GeneNumber_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_B_GeneNumber_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_GeneNumber_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_C_GeneNumber_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_GeneNumber_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_D_GeneNumber_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)

    stats_file_A_TotalDepth_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_A_TotalDepth_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_B_TotalDepth_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_B_TotalDepth_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_C_TotalDepth_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_C_TotalDepth_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)
    stats_file_D_TotalDepth_pct_by_all = '%s/%s_KEGG_wd/%s_ko_stats_D_TotalDepth_pct_by_all.txt'    % (op_dir, in_file_basename, in_file_basename)

    ################################################# parse blast results ##################################################

    if run_blast is True:

        # store blast results in dict
        query_to_db_seq_dict = {}
        for each_query in open(blast_results_best_hit):
            each_query_split = each_query.strip().split('\t')
            query_id = each_query_split[0]
            db_seq = each_query_split[1]
            query_to_db_seq_dict[query_id] = db_seq

        # get all query sequence id
        query_seq_id_list = []
        for each_seq in SeqIO.parse(pwd_input_file, 'fasta'):
            query_seq_id_list.append(str(each_seq.id))

        # get ko id at level D for all query genes
        KO_assignment_file_handle = open(KO_assignment_file_D, 'w')
        for each_query_seq in sorted(query_seq_id_list):

            if each_query_seq in query_to_db_seq_dict:
                db_hit_id = query_to_db_seq_dict[each_query_seq]

                if db_hit_id in db_seq_to_KO_dict:
                    db_hit_id_ko = db_seq_to_KO_dict[db_hit_id]
                    if ',' in db_hit_id_ko:
                        db_hit_id_ko_split = db_hit_id_ko.split(',')
                        for each_db_hit_id_ko in db_hit_id_ko_split:
                            KO_assignment_file_handle.write('%s\t%s\n' % (each_query_seq, each_db_hit_id_ko))
                    else:
                        KO_assignment_file_handle.write('%s\t%s\n' % (each_query_seq, db_hit_id_ko))
                else:
                    KO_assignment_file_handle.write('%s\n' % (each_query_seq))
            else:
                KO_assignment_file_handle.write('%s\n' % (each_query_seq))
        KO_assignment_file_handle.close()

    else:
        KO_assignment_file_D = pwd_input_file

    # get ko id at all levels for all query genes
    ko_assign_ABCD_handle = open(KO_assignment_file_DCBA, 'w')
    ko_assign_ABCD_handle.write('Gene_id\tko_A\tko_B\tko_C\tko_D\tDesc_A\tDesc_B\tDesc_C\tDesc_D\n')
    query_seq_id_all = set()
    genes_with_ko = set()
    for query_gene in open(KO_assignment_file_D):
        query_gene_split = query_gene.strip().split('\t')
        gene_ID = query_gene_split[0]

        if len(query_gene_split) == 1:
            query_seq_id_all.add(query_gene_split[0])
            ko_assign_ABCD_handle.write('%s\n' % gene_ID)

        if len(query_gene_split) == 2:
            query_seq_id_all.add(query_gene_split[0])
            genes_with_ko.add(query_gene_split[0])
            KO_ID = query_gene_split[1]
            if KO_ID in D2ABCD_dict:
                KO_ID_ABCD = D2ABCD_dict[KO_ID]

                if len(KO_ID_ABCD) == 1:
                    KO_DCBA_list = KO_ID_ABCD[0].split('|')[::-1]
                    KO_DCBA_list_only_id = [i.split('_')[1] for i in KO_DCBA_list]
                    desc_A = As_description_dict[KO_DCBA_list_only_id[3]]
                    desc_B = Bs_description_dict[KO_DCBA_list_only_id[2]]
                    desc_C = Cs_description_dict[KO_DCBA_list_only_id[1]]
                    desc_D = Ds_description_dict[KO_DCBA_list_only_id[0]]
                    ko_assign_ABCD_handle.write('%s\t%s\t%s\t%s\t%s\t%s\n' % (gene_ID,
                                                                             '\t'.join(KO_DCBA_list[::-1]),
                                                                             desc_A, desc_B, desc_C, desc_D))

                if len(KO_ID_ABCD) > 1:
                    for each_ABCD in KO_ID_ABCD:
                        each_KO_DCBA_list = each_ABCD.split('|')[::-1]
                        each_KO_DCBA_list_only_id = [i.split('_')[1] for i in each_KO_DCBA_list]
                        each_desc_A = As_description_dict[each_KO_DCBA_list_only_id[3]]
                        each_desc_B = Bs_description_dict[each_KO_DCBA_list_only_id[2]]
                        each_desc_C = Cs_description_dict[each_KO_DCBA_list_only_id[1]]
                        each_desc_D = Ds_description_dict[each_KO_DCBA_list_only_id[0]]
                        ko_assign_ABCD_handle.write('%s\t%s\t%s\t%s\t%s\t%s\n' % (gene_ID,
                                                                                 '\t'.join(each_KO_DCBA_list[::-1]),
                                                                                 each_desc_A, each_desc_B,
                                                                                 each_desc_C, each_desc_D))

    ko_assign_ABCD_handle.close()

    ##################################################### Get summary ######################################################

    # read in depth info
    gene_depth_dict = {}
    if depth_file is not None:
        for each_depth in open(depth_file):
            each_depth_split = each_depth.strip().split('\t')
            gene_depth_dict[each_depth_split[0]] = float(each_depth_split[1])

    # get total number and depth of all genes in one file
    total_depth_for_all_query_genes = 0
    genes_with_ko_TotalDepth = 0
    if depth_file is not None:
        for gene in query_seq_id_all:
            gene_depth = gene_depth_dict[gene]
            total_depth_for_all_query_genes += gene_depth

        genes_with_ko_TotalDepth = get_gene_list_TotalDepth(genes_with_ko, gene_depth_dict)

    identified_ko_A_list = []
    identified_ko_B_list = []
    identified_ko_C_list = []
    identified_ko_D_list = []
    ko_A_to_gene_member_dict = {}
    ko_B_to_gene_member_dict = {}
    ko_C_to_gene_member_dict = {}
    ko_D_to_gene_member_dict = {}
    ko_NA_to_gene_member_list = []
    for each_query in open(KO_assignment_file_DCBA):
        if not each_query.startswith('Gene_id'):
            each_query_split = each_query.strip().split('\t')
            query_id = each_query_split[0]

            if len(each_query_split) == 1:
                ko_NA_to_gene_member_list.append(query_id)

            if len(each_query_split) > 1:
                query_ko_A = each_query_split[1]
                query_ko_B = each_query_split[2]
                query_ko_C = each_query_split[3]
                query_ko_D = each_query_split[4]

                if query_ko_A not in identified_ko_A_list:
                    identified_ko_A_list.append(query_ko_A)
                if query_ko_B not in identified_ko_B_list:
                    identified_ko_B_list.append(query_ko_B)
                if query_ko_C not in identified_ko_C_list:
                    identified_ko_C_list.append(query_ko_C)
                if query_ko_D not in identified_ko_D_list:
                    identified_ko_D_list.append(query_ko_D)

                if query_ko_A not in ko_A_to_gene_member_dict:
                    ko_A_to_gene_member_dict[query_ko_A] = [query_id]
                else:
                    if query_id not in ko_A_to_gene_member_dict[query_ko_A]:
                        ko_A_to_gene_member_dict[query_ko_A].append(query_id)

                if query_ko_B not in ko_B_to_gene_member_dict:
                    ko_B_to_gene_member_dict[query_ko_B] = [query_id]
                else:
                    if query_id not in ko_B_to_gene_member_dict[query_ko_B]:
                        ko_B_to_gene_member_dict[query_ko_B].append(query_id)

                if query_ko_C not in ko_C_to_gene_member_dict:
                    ko_C_to_gene_member_dict[query_ko_C] = [query_id]
                else:
                    if query_id not in ko_C_to_gene_member_dict[query_ko_C]:
                        ko_C_to_gene_member_dict[query_ko_C].append(query_id)

                if query_ko_D not in ko_D_to_gene_member_dict:
                    ko_D_to_gene_member_dict[query_ko_D] = [query_id]
                else:
                    if query_id not in ko_D_to_gene_member_dict[query_ko_D]:
                        ko_D_to_gene_member_dict[query_ko_D].append(query_id)

    #################### write out GeneNumber and TotalDepth stats ####################

    write_out_stats_GeneNumber(identified_ko_A_list, ko_A_to_gene_member_dict, As_description_dict, stats_file_A_GeneNumber)
    write_out_stats_GeneNumber(identified_ko_B_list, ko_B_to_gene_member_dict, Bs_description_dict, stats_file_B_GeneNumber)
    write_out_stats_GeneNumber(identified_ko_C_list, ko_C_to_gene_member_dict, Cs_description_dict, stats_file_C_GeneNumber)
    write_out_stats_GeneNumber(identified_ko_D_list, ko_D_to_gene_member_dict, Ds_description_dict, stats_file_D_GeneNumber)
    if depth_file is not None:
        write_out_stats_TotalDepth(identified_ko_A_list, ko_A_to_gene_member_dict, gene_depth_dict, As_description_dict, stats_file_A_TotalDepth)
        write_out_stats_TotalDepth(identified_ko_B_list, ko_B_to_gene_member_dict, gene_depth_dict, Bs_description_dict, stats_file_B_TotalDepth)
        write_out_stats_TotalDepth(identified_ko_C_list, ko_C_to_gene_member_dict, gene_depth_dict, Cs_description_dict, stats_file_C_TotalDepth)
        write_out_stats_TotalDepth(identified_ko_D_list, ko_D_to_gene_member_dict, gene_depth_dict, Ds_description_dict, stats_file_D_TotalDepth)

    #################### write out GeneNumber and TotalDepth stats (pct) ####################

    AnnotateNorm(stats_file_A_GeneNumber, True, 2, len(genes_with_ko), stats_file_A_GeneNumber_pct, 'KO\tGeneNumber_pct\tDescription\n')
    AnnotateNorm(stats_file_B_GeneNumber, True, 2, len(genes_with_ko), stats_file_B_GeneNumber_pct, 'KO\tGeneNumber_pct\tDescription\n')
    AnnotateNorm(stats_file_C_GeneNumber, True, 2, len(genes_with_ko), stats_file_C_GeneNumber_pct, 'KO\tGeneNumber_pct\tDescription\n')
    AnnotateNorm(stats_file_D_GeneNumber, True, 2, len(genes_with_ko), stats_file_D_GeneNumber_pct, 'KO\tGeneNumber_pct\tDescription\n')
    if depth_file is not None:
        AnnotateNorm(stats_file_A_TotalDepth, True, 2, genes_with_ko_TotalDepth, stats_file_A_TotalDepth_pct, 'KO\tTotalDepth_pct\tDescription\n')
        AnnotateNorm(stats_file_B_TotalDepth, True, 2, genes_with_ko_TotalDepth, stats_file_B_TotalDepth_pct, 'KO\tTotalDepth_pct\tDescription\n')
        AnnotateNorm(stats_file_C_TotalDepth, True, 2, genes_with_ko_TotalDepth, stats_file_C_TotalDepth_pct, 'KO\tTotalDepth_pct\tDescription\n')
        AnnotateNorm(stats_file_D_TotalDepth, True, 2, genes_with_ko_TotalDepth, stats_file_D_TotalDepth_pct, 'KO\tTotalDepth_pct\tDescription\n')

    #################### write out GeneNumber and TotalDepth stats (pct_by_all) ####################

    if pct_by_all is True:
        AnnotateNorm(stats_file_A_GeneNumber, True, 2, len(query_seq_id_all), stats_file_A_GeneNumber_pct_by_all, 'KO\tGeneNumber_pct_by_all\tDescription\n')
        AnnotateNorm(stats_file_B_GeneNumber, True, 2, len(query_seq_id_all), stats_file_B_GeneNumber_pct_by_all, 'KO\tGeneNumber_pct_by_all\tDescription\n')
        AnnotateNorm(stats_file_C_GeneNumber, True, 2, len(query_seq_id_all), stats_file_C_GeneNumber_pct_by_all, 'KO\tGeneNumber_pct_by_all\tDescription\n')
        AnnotateNorm(stats_file_D_GeneNumber, True, 2, len(query_seq_id_all), stats_file_D_GeneNumber_pct_by_all, 'KO\tGeneNumber_pct_by_all\tDescription\n')
        if depth_file is not None:
            AnnotateNorm(stats_file_A_TotalDepth, True, 2, total_depth_for_all_query_genes, stats_file_A_TotalDepth_pct_by_all, 'KO\tTotalDepth_pct_by_all\tDescription\n')
            AnnotateNorm(stats_file_B_TotalDepth, True, 2, total_depth_for_all_query_genes, stats_file_B_TotalDepth_pct_by_all, 'KO\tTotalDepth_pct_by_all\tDescription\n')
            AnnotateNorm(stats_file_C_TotalDepth, True, 2, total_depth_for_all_query_genes, stats_file_C_TotalDepth_pct_by_all, 'KO\tTotalDepth_pct_by_all\tDescription\n')
            AnnotateNorm(stats_file_D_TotalDepth, True, 2, total_depth_for_all_query_genes, stats_file_D_TotalDepth_pct_by_all, 'KO\tTotalDepth_pct_by_all\tDescription\n')


def get_KEGG_annot_df(annotation_dir, stats_level, annotation_df_absolute_num, annotation_df_pct, annotation_df_pct_by_all, with_depth, pct_by_all):

    annotation_dir_re = '%s/*_KEGG_wd' % annotation_dir
    annotation_folder_list = [os.path.basename(file_name) for file_name in glob.glob(annotation_dir_re)]

    ko_num_dict = {}
    ko_num_pct_dict = {}
    ko_num_pct_by_all_dict = {}
    all_identified_ko = set()
    for annotation_folder in annotation_folder_list:

        annotation_folder_basename = annotation_folder.split('_KEGG_wd')[0]

        if with_depth is False:
            pwd_annotation_stats_file =             '%s/%s/%s_ko_stats_%s_GeneNumber.txt'               % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)
            pwd_annotation_stats_file_pct =         '%s/%s/%s_ko_stats_%s_GeneNumber_pct.txt'           % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)
            pwd_annotation_stats_file_pct_by_all =  '%s/%s/%s_ko_stats_%s_GeneNumber_pct_by_all.txt'    % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)

        else:
            pwd_annotation_stats_file =             '%s/%s/%s_ko_stats_%s_TotalDepth.txt'               % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)
            pwd_annotation_stats_file_pct =         '%s/%s/%s_ko_stats_%s_TotalDepth_pct.txt'           % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)
            pwd_annotation_stats_file_pct_by_all =  '%s/%s/%s_ko_stats_%s_TotalDepth_pct_by_all.txt'    % (annotation_dir, annotation_folder, annotation_folder_basename, stats_level)

        current_ko_to_num_dict = {}
        for ko in open(pwd_annotation_stats_file):
            if not ko.startswith('KO\t'):
                ko_split = ko.strip().split('\t')
                if with_depth is False:
                    current_ko_to_num_dict[ko_split[0]] = int(ko_split[1])
                else:
                    current_ko_to_num_dict[ko_split[0]] = float(ko_split[1])
                all_identified_ko.add(ko_split[0])

        current_ko_to_num_pct_dict = {}
        for ko in open(pwd_annotation_stats_file_pct):
            if not ko.startswith('KO\t'):
                ko_split = ko.strip().split('\t')
                current_ko_to_num_pct_dict[ko_split[0]] = float(ko_split[1])
                all_identified_ko.add(ko_split[0])

        if pct_by_all is True:
            current_ko_to_num_pct_by_all_dict = {}
            for ko in open(pwd_annotation_stats_file_pct_by_all):
                if not ko.startswith('KO\t'):
                    ko_split = ko.strip().split('\t')
                    current_ko_to_num_pct_by_all_dict[ko_split[0]] = float(ko_split[1])
                    all_identified_ko.add(ko_split[0])

        ko_num_dict[annotation_folder_basename] = current_ko_to_num_dict
        ko_num_pct_dict[annotation_folder_basename] = current_ko_to_num_pct_dict
        if pct_by_all is True:
            ko_num_pct_by_all_dict[annotation_folder_basename] = current_ko_to_num_pct_by_all_dict

    all_identified_ko_list = sorted([i for i in all_identified_ko])

    annotation_df_absolute_num_handle = open(annotation_df_absolute_num, 'w')
    annotation_df_absolute_num_handle.write('\t%s\n' % '\t'.join(all_identified_ko_list))
    annotation_df_percentage_handle = open(annotation_df_pct, 'w')
    annotation_df_percentage_handle.write('\t%s\n' % '\t'.join(all_identified_ko_list))
    if pct_by_all is True:
        annotation_df_percentage_by_all_handle = open(annotation_df_pct_by_all, 'w')
        annotation_df_percentage_by_all_handle.write('\t%s\n' % '\t'.join(all_identified_ko_list))
    for annotation_folder in sorted(annotation_folder_list):

        annotation_folder_basename = annotation_folder.split('_KEGG_wd')[0]
        current_ko_num_dict = ko_num_dict[annotation_folder_basename]
        current_ko_num_dict_pct = ko_num_pct_dict[annotation_folder_basename]
        if pct_by_all is True:
            current_ko_num_dict_pct_by_all = ko_num_pct_by_all_dict[annotation_folder_basename]

        current_ko_num_list = []
        current_ko_num_list_pct = []
        current_ko_num_list_pct_by_all = []
        for identified_ko in all_identified_ko_list:

            # get num list
            identified_ko_num = 0
            identified_ko_num_pct = 0
            identified_ko_num_pct_by_all = 0
            if identified_ko in current_ko_num_dict:
                identified_ko_num = current_ko_num_dict[identified_ko]
                identified_ko_num_pct = current_ko_num_dict_pct[identified_ko]
                if pct_by_all is True:
                    identified_ko_num_pct_by_all = current_ko_num_dict_pct_by_all[identified_ko]

            current_ko_num_list.append(identified_ko_num)
            current_ko_num_list_pct.append(identified_ko_num_pct)
            if pct_by_all is True:
                current_ko_num_list_pct_by_all.append(identified_ko_num_pct_by_all)

        # write out
        annotation_df_absolute_num_handle.write('%s\t%s\n' % (annotation_folder_basename, '\t'.join([str(i) for i in current_ko_num_list])))
        annotation_df_percentage_handle.write('%s\t%s\n' % (annotation_folder_basename, '\t'.join([str(i) for i in current_ko_num_list_pct])))
        if pct_by_all is True:
            annotation_df_percentage_by_all_handle.write('%s\t%s\n' % (annotation_folder_basename, '\t'.join([str(i) for i in current_ko_num_list_pct_by_all])))

    annotation_df_absolute_num_handle.close()
    annotation_df_percentage_handle.close()
    if pct_by_all is True:
        annotation_df_percentage_by_all_handle.close()


def Annotation_KEGG(args):

    input_file_faa =      args['seq_in']
    input_file_user_ko =  args['ko_in']
    file_extension =      args['x']
    depth_file =          args['depth']
    pct_by_all =          args['pct_by_all']
    KEGG_DB_folder =      args['db_dir']
    run_diamond =         args['diamond']
    num_threads =         args['t']
    evalue_cutoff =       args['evalue']


    run_blast = None
    if (input_file_faa is not None) and (input_file_user_ko is None):
        run_blast = True
    elif (input_file_faa is None) and (input_file_user_ko is not None):
        run_blast = False
    else:
        print(datetime.now().strftime(time_format) + 'Please provide input file with either "-seq_in" or "-ko_in", do not provide both')
        exit()

    if run_blast is True:
        input_file_folder = input_file_faa
    else:
        input_file_folder = input_file_user_ko

    # check whether input file/folder exist
    if (os.path.isfile(input_file_folder) is False) and (os.path.isdir(input_file_folder) is False):
        print(datetime.now().strftime(time_format) + 'input file/folder not found, program exited')
        exit()

    if run_blast is True:
        print(datetime.now().strftime(time_format) + 'Input sequence file detected, will run blastp/diamond first')
        sleep(0.5)
    else:
        print(datetime.now().strftime(time_format) + 'Annotation results provided, blastp/diamond skipped')
        sleep(0.5)


    ################################################# define file name #################################################

    KEGG_DB_seq =         '%s/kegg_db_seq.fasta'            % KEGG_DB_folder
    KEGG_DB_seq_diamond = '%s/kegg_db_seq.fasta.dmnd'       % KEGG_DB_folder
    KEGG_DB_seq2ko =      '%s/kegg_db_seq2ko.txt'           % KEGG_DB_folder
    KEGG_DB_ko =          '%s/ko00001.keg'                  % KEGG_DB_folder


    ########################################## check whether diamond db exist ##########################################

    if (run_blast is True) and (run_diamond is True):
        if os.path.isfile(KEGG_DB_seq_diamond) is False:
            print(datetime.now().strftime(time_format) + 'DB file not found, making diamond db with %s' % KEGG_DB_seq)

            if os.path.isfile(KEGG_DB_seq) is True:
                diamond_makedb_cmd = 'diamond makedb --in %s --db %s --quiet' % (KEGG_DB_seq, KEGG_DB_seq_diamond)
                os.system(diamond_makedb_cmd)
            else:
                print(datetime.now().strftime(time_format) + '%s not found, program exited' % KEGG_DB_seq)
                exit()

    ########################################### check whether blast+ db exist ##########################################

    if (run_blast is True) and (run_diamond is False):

        unfound_db_index_file = []
        for db_index in ['phr', 'pin', 'pnd', 'pni', 'pog', 'psd', 'psi', 'psq']:
            pwd_db_index = '%s/kegg_db_seq.fasta.%s' % (KEGG_DB_folder, db_index)
            if not os.path.isfile(pwd_db_index):
                unfound_db_index_file.append(db_index)
        if len(unfound_db_index_file) > 0:
            print(datetime.now().strftime(time_format) + 'blast db index not found, runing makeblastdb first')
            makeblastdb_cmd = 'makeblastdb -in %s -dbtype prot -parse_seqids -logfile %s.log' % (KEGG_DB_seq, KEGG_DB_seq)
            os.system(makeblastdb_cmd)
            print(datetime.now().strftime(time_format) + 'makeblastdb finished')


    ######################################### Run blastp with multiprocessing ##########################################

    # check whether the input file is a file or folder
    if os.path.isfile(input_file_folder) is True:
        input_file_path, input_file_basename, input_file_ext = sep_path_basename_ext(input_file_folder)
        run_blast_worker([input_file_folder, run_blast, run_diamond, KEGG_DB_seq, KEGG_DB_seq_diamond, input_file_path, evalue_cutoff, num_threads])

    if os.path.isdir(input_file_folder) is True:

        # create output folder
        output_folder = '%s_KEGG_wd' % input_file_folder
        force_create_folder(output_folder)

        # check whether input genome exist
        input_file_re = '%s/*.%s' % (input_file_folder, file_extension)
        input_file_name_list = [os.path.basename(file_name) for file_name in glob.glob(input_file_re)]

        if len(input_file_name_list) == 0:
            print(datetime.now().strftime(time_format) + 'input file not found, program exited')
            exit()

        # run blastp with multiprocessing
        if run_blast is True:
            print(datetime.now().strftime(time_format) + 'Running Blast/Diamond for %s input files with %s cores' % (len(input_file_name_list), num_threads))

        list_for_multiple_arguments_blast = []
        for input_file in input_file_name_list:
            pwd_input_file = '%s/%s' % (input_file_folder, input_file)
            list_for_multiple_arguments_blast.append([pwd_input_file, run_blast, run_diamond, KEGG_DB_seq, KEGG_DB_seq_diamond, output_folder, evalue_cutoff, 1])

        # run blastp with multiprocessing
        pool = mp.Pool(processes=num_threads)
        pool.map(run_blast_worker, list_for_multiple_arguments_blast)
        pool.close()
        pool.join()

    ############################################## Read in KEGG DB files ###############################################

    print(datetime.now().strftime(time_format) + 'Read in KEGG DB files')

    As_description_dict = {}
    Bs_description_dict = {}
    Cs_description_dict = {}
    Ds_description_dict = {}
    D2ABCD_dict = {}
    current_A = ''
    current_B = ''
    current_C = ''
    for each_line in open(KEGG_DB_ko):
        if each_line[0] in ['A', 'B', 'C', 'D']:
            each_line_split = each_line.strip().split(' ')

            if each_line[0] == 'A':
                current_A_id = each_line_split[0]
                current_A_description = ' '.join(each_line_split[1:])
                current_A = current_A_id
                As_description_dict[current_A_id] = current_A_description

            elif each_line[0] == 'B':
                if len(each_line_split) > 1:
                    current_B_id = each_line_split[2]
                    current_B_description = ' '.join(each_line_split[3:])
                    current_B = current_B_id
                    Bs_description_dict[current_B_id] = current_B_description

            elif each_line[0] == 'C':
                current_C_id = each_line_split[4]
                current_C_description = ' '.join(each_line_split[5:])
                current_C = current_C_id
                Cs_description_dict[current_C_id] = current_C_description

            elif each_line[0] == 'D':
                current_D_id = each_line_split[6]
                current_D_description = ' '.join(each_line_split[7:])
                Ds_description_dict[current_D_id] = current_D_description
                ABCD_value = 'A_%s|B_%s|C_%s|D_%s' % (current_A, current_B, current_C, current_D_id)
                if current_D_id not in D2ABCD_dict:
                    D2ABCD_dict[current_D_id] = [ABCD_value]
                elif (current_D_id in D2ABCD_dict) and (ABCD_value not in D2ABCD_dict[current_D_id]):
                    D2ABCD_dict[current_D_id].append(ABCD_value)

    # get db_seq_to_KO_dict
    db_seq_to_KO_dict = {}
    if run_blast is True:
        for each_hit in open(KEGG_DB_seq2ko):
            each_hit_split = each_hit.strip().split('\t')
            db_seq = each_hit_split[0]
            hit_id_KO = each_hit_split[1]
            if hit_id_KO != '':
                db_seq_to_KO_dict[db_seq] = hit_id_KO

    ########################################################################################################################

    # check whether the input file is a file or folder
    if os.path.isfile(input_file_folder) is True:

        # check whether depth file exist
        if depth_file is not None:
            if os.path.isfile(depth_file) is False:
                print(datetime.now().strftime(time_format) + 'specified depth file not found, program exited!')
                exit()

        print(datetime.now().strftime(time_format) + 'Running KEGG annotation for 1 file with %s cores' % (num_threads))
        input_file_path, input_file_basename, input_file_ext = sep_path_basename_ext(input_file_folder)
        parse_blast_op_worker([input_file_folder, run_blast, As_description_dict, Bs_description_dict, Cs_description_dict, Ds_description_dict, D2ABCD_dict, db_seq_to_KO_dict, input_file_path, depth_file, pct_by_all])


    if os.path.isdir(input_file_folder) is True:

        input_file_re = '%s/*.%s' % (input_file_folder, file_extension)
        input_file_name_list = [os.path.basename(file_name) for file_name in glob.glob(input_file_re)]

        # check whether depth file exist
        if depth_file is not None:

            if os.path.isfile(depth_file) is True:
                print(datetime.now().strftime(
                    time_format) + 'please provide the folder containing individual depth files (with extension .depth) for each of your input sequence file.')
                print(datetime.now().strftime(time_format) + 'single depth file (not folder) detected, program exited!')
                exit()

            if os.path.isdir(depth_file) is False:
                print(datetime.now().strftime(time_format) + 'specified depth folder not found, program exited!')
                exit()

            if os.path.isdir(depth_file) is True:

                undetected_depth_file = []
                for input_seq_file in input_file_name_list:
                    input_seq_file_basename = '.'.join(input_seq_file.split('.')[:-1])
                    input_seq_file_depth = '%s/%s.depth' % (depth_file, input_seq_file_basename)
                    if os.path.isfile(input_seq_file_depth) is False:
                        undetected_depth_file.append(input_seq_file_depth)

                if len(undetected_depth_file) > 0:
                    print(datetime.now().strftime(time_format) + 'the following depth files not found, program exited!')
                    print(','.join(undetected_depth_file))
                    exit()

        # create output folder
        output_folder = '%s_KEGG_wd' % input_file_folder
        input_folder_name = input_file_folder
        if '/' in input_file_folder:
            input_folder_name = input_file_folder.split('/')[-1]

        # parse blast results with multiprocessing
        if run_blast is True:
            print(datetime.now().strftime(time_format) + 'Parsing Blast/Diamond results for %s input files with %s cores' % (len(input_file_name_list), num_threads))

        list_for_multiple_arguments_parse_blast_op = []
        for input_file in input_file_name_list:

            input_file_basename = '.'.join(input_file.split('.')[:-1])
            pwd_input_file = '%s/%s' % (input_file_folder, input_file)

            # get path to current depth file
            if depth_file is None:
                input_file_depth = None
            else:
                input_file_depth = '%s/%s.depth' % (depth_file, input_file_basename)

            list_for_multiple_arguments_parse_blast_op.append([pwd_input_file, run_blast, As_description_dict, Bs_description_dict, Cs_description_dict, Ds_description_dict, D2ABCD_dict, db_seq_to_KO_dict, output_folder, input_file_depth, pct_by_all])

        # parse blast results with multiprocessing
        pool = mp.Pool(processes=num_threads)
        pool.map(parse_blast_op_worker, list_for_multiple_arguments_parse_blast_op)
        pool.close()
        pool.join()

        ######################################################### get dataframe #########################################################

        print(datetime.now().strftime(time_format) + 'Data matrix exported to:')

        for ko_level in ['A', 'B', 'C', 'D']:
            annotation_df_GeneNumber =            '%s/%s_%s_GeneNumber.txt'            % (output_folder, input_folder_name, ko_level)
            annotation_df_GeneNumber_pct =        '%s/%s_%s_GeneNumber_pct.txt'        % (output_folder, input_folder_name, ko_level)
            annotation_df_GeneNumber_pct_by_all = '%s/%s_%s_GeneNumber_pct_by_all.txt' % (output_folder, input_folder_name, ko_level)
            annotation_df_TotalDepth =            '%s/%s_%s_TotalDepth.txt'            % (output_folder, input_folder_name, ko_level)
            annotation_df_TotalDepth_pct =        '%s/%s_%s_TotalDepth_pct.txt'        % (output_folder, input_folder_name, ko_level)
            annotation_df_TotalDepth_pct_by_all = '%s/%s_%s_TotalDepth_pct_by_all.txt' % (output_folder, input_folder_name, ko_level)

            #################### get GeneNumber df and report ####################

            get_KEGG_annot_df(output_folder, ko_level, annotation_df_GeneNumber, annotation_df_GeneNumber_pct, annotation_df_GeneNumber_pct_by_all, with_depth=False, pct_by_all=pct_by_all)

            print(annotation_df_GeneNumber.split('/')[-1])
            print(annotation_df_GeneNumber_pct.split('/')[-1])
            if pct_by_all is True:
                print(annotation_df_GeneNumber_pct_by_all.split('/')[-1])

            #################### get TotalDepth df and report ####################

            if depth_file is not None:
                get_KEGG_annot_df(output_folder, ko_level, annotation_df_TotalDepth, annotation_df_TotalDepth_pct, annotation_df_TotalDepth_pct_by_all, with_depth=True, pct_by_all=pct_by_all)

                print(annotation_df_TotalDepth.split('/')[-1])
                print(annotation_df_TotalDepth_pct.split('/')[-1])
                if pct_by_all is True:
                    print(annotation_df_TotalDepth_pct_by_all.split('/')[-1])

    ################################################## Final report ####################################################

    print(datetime.now().strftime(time_format) + 'Done!')


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('-seq_in',      required=False,                            help='faa file')
    parser.add_argument('-ko_in',       required=False,                            help='annotation results from BlastKOALA/GhostKOALA, normally with name user_ko.txt')
    parser.add_argument('-x',           required=False,                            help='file extension')
    parser.add_argument('-depth',       required=False, default=None,              help='gene depth file/folder')
    parser.add_argument('-pct_by_all',  required=False, action='store_true',       help='normalize by all query genes, rather than those with ko assignment')
    parser.add_argument('-db_dir',      required=True,                             help='folder holds sequence, seq2ko and ko00001.keg files')
    parser.add_argument('-diamond',     required=False, action='store_true',       help='run diamond (for big dataset), default is NCBI blastp')
    parser.add_argument('-t',           required=False, default=1,     type=int,   help='number of threads, default: 1')
    parser.add_argument('-evalue',      required=False, default=0.001, type=float, help='evalue cutoff, default: 0.001')

    args = vars(parser.parse_args())
    Annotation_KEGG(args)
